import logging
from datetime import timedelta
from fractions import Fraction
from os import PathLike, fspath
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator, Sequence, Tuple, Union

import numpy as np
from typing_extensions import Self

if TYPE_CHECKING:
    from av import VideoFrame

# if __debug__:
#     import av
#     import cv2

logger = logging.getLogger(__name__)


def _raise_exists(outpath: Path, overwrite: bool = False) -> None:
    if not overwrite and outpath.exists():
        raise FileExistsError()


class NoGoodFrame(Exception):
    pass


class NoKeyFrame(Exception):
    pass


class BadFile(Exception):
    pass


class VideoBase:
    single_frame_ratio = 0.5  # fixme: what is this used for?
    time_base: Fraction
    native_duration: int

    def calculate_offsets(self, time_base: Fraction, duration: int) -> Iterator[int]:
        raise NotImplementedError

    def _frame_range(self, time_base: Fraction, duration: int) -> Iterator[int]:
        raise NotImplementedError

    def frame_range(self) -> Iterator[int]:
        return self._frame_range(self.time_base, self.native_duration)

    def _get_frame(self, offset: int, native: bool = False) -> Tuple[float, np.ndarray]:
        raise NotImplementedError

    def _frame_to_file(self, frame: Any, outpath: PathLike) -> None:
        raise NotImplementedError

    def iterate(self) -> Iterator[Tuple[float, np.ndarray]]:
        for offset in self.calculate_offsets(self.time_base, self.native_duration):
            try:
                yield self._get_frame(offset)
            except NoKeyFrame as e:
                yield offset, e

    def iterall(self, native: bool = False) -> Iterator[Tuple[float, np.ndarray]]:
        raise NotImplementedError

    def save_frame_to_file(self, pos: float, outpath: PathLike) -> None:
        frametime, frame = self._get_frame(int(self.native_duration * pos), native=True)
        self._frame_to_file(frame, outpath)

    def close(self) -> None:
        raise NotImplementedError

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def fourcc(integer):
    import struct

    return struct.pack("<i", integer)

    # return integer.to_bytes(4, "little") # python 3 only


class CvVideo(VideoBase):
    def __init__(self, path: Union[str, PathLike, int]) -> None:
        self.cv2 = self.import_backend()

        if isinstance(path, PathLike):
            path = fspath(path)

        self.cap = self.cv2.VideoCapture(path)
        logger.debug("Reading video using backend: %s", self.cap.getBackendName())

        frame_width = int(self.cap.get(self.cv2.CAP_PROP_FRAME_WIDTH))
        frame_height = int(self.cap.get(self.cv2.CAP_PROP_FRAME_HEIGHT))
        frame_count = int(self.cap.get(self.cv2.CAP_PROP_FRAME_COUNT))

        if frame_width == 0 or frame_height == 0:
            raise ValueError(f"Invalid video file (zero frame dimensions): {path}")

        if frame_count == 0:
            raise ValueError(f"Invalid video file (zero frame count): {path}")

        pixel_aspect_ratio = Fraction(
            int(self.cap.get(self.cv2.CAP_PROP_SAR_NUM)), int(self.cap.get(self.cv2.CAP_PROP_SAR_DEN))
        )

        if pixel_aspect_ratio == 0:
            pixel_aspect_ratio = Fraction(1, 1)

        display_aspect_ratio = pixel_aspect_ratio * Fraction(frame_width, frame_height)

        fps = self.cap.get(self.cv2.CAP_PROP_FPS)
        try:
            duration = timedelta(seconds=frame_count / fps)
        except ZeroDivisionError:
            raise BadFile(f"Cannot open {path}")

        self.native_duration = frame_count  # exclusive
        self.time_base = Fraction(1 / fps)

        self.meta = {
            "width": frame_width,
            "height": frame_height,
            "duration": duration,  # container duration in seconds
            "fps": fps,
            "sample_aspect_ratio": pixel_aspect_ratio,
            "display_aspect_ratio": display_aspect_ratio,
            "format": fourcc(int(self.cap.get(self.cv2.CAP_PROP_CODEC_PIXEL_FORMAT))).decode("ascii"),
        }

    @staticmethod
    def import_backend():
        import cv2

        return cv2

    def iterall(self, native: bool = False) -> Iterator[Tuple[float, np.ndarray]]:
        while True:
            retval, image = self.cap.read()
            offset = self.cap.get(self.cv2.CAP_PROP_POS_FRAMES)
            if retval:
                if not native:
                    image = self.cv2.cvtColor(image, self.cv2.COLOR_BGR2RGB)
                yield self.time_to_seconds(offset), image
            else:
                break

    def show(self, title: str = "iter_video") -> Iterator[np.ndarray]:
        """Show video. Quit using key 'q'"""

        try:
            for _time, image in self.iterall(native=True):
                yield image
                self.cv2.imshow(title, image)
                if self.cv2.waitKey(1) & 0xFF == ord("q"):
                    return
        finally:
            self.cv2.destroyAllWindows()

    def time_to_seconds(self, offset: int) -> float:
        return offset / self.meta["fps"]

    def _get_frame(self, offset: int, native: bool = False) -> Tuple[float, np.ndarray]:
        self.cap.set(self.cv2.CAP_PROP_POS_FRAMES, offset)

        ret, frame = self.cap.read()
        if ret:
            if not native:
                frame = self.cv2.cvtColor(frame, self.cv2.COLOR_BGR2RGB)
            return self.time_to_seconds(offset), frame
        else:
            raise NoGoodFrame("Could not find good frame")

    def _frame_to_file(self, frame: np.ndarray, outpath: Path) -> None:
        is_success: bool
        im_buf: np.ndarray

        is_success, im_buf = self.cv2.imencode(outpath.suffix, frame)
        assert is_success
        im_buf.tofile(fspath(outpath))  # cv2.imwrite doesn't handle unicode paths correctly

    def close(self) -> None:
        self.cap.release()


def object_attributes(obj):
    def inner():
        for attr in dir(obj):
            val = getattr(obj, attr)
            if not callable(val):
                yield attr, val

    return dict(inner())


class AvVideo(VideoBase):
    vcc_attrs = [
        "coded_height",
        "coded_width",
        "display_aspect_ratio",
        "encoded_frame_count",
        "format",
        "framerate",
        "gop_size",
        "has_b_frames",
        "height",
        "pix_fmt",
        "reformatter",
        "sample_aspect_ratio",
        "width",
    ]

    def __init__(self, path: PathLike, videostream: int = 0) -> None:
        self.av = self.import_backend()

        if isinstance(path, PathLike):
            path = fspath(path)

        try:
            self.container = self.av.open(path, "r")
        except self.av.error.InvalidDataError:
            raise BadFile(f"Cannot open {path}")

        # why was this here in the first place?
        # if self.container.format.name == "matroska,webm":
        #     raise BadFile("Matroska files are currently not supported :(")

        self.vstream = self.container.streams.video[videostream]
        self.vstream.thread_type = "AUTO"

        duration = timedelta(seconds=self.container.duration / self.av.time_base)

        vcc = self.vstream.codec_context
        vcc = {attr: getattr(vcc, attr) for attr in self.vcc_attrs}
        del vcc["format"]

        vcc["average_rate"] = getattr(self.vstream, "average_rate", None)  # for av>=10.0.0

        self.native_duration = self.vstream.duration  # exclusive
        self.time_base = self.vstream.time_base

        if not self.native_duration:  # if stream duration is not available, use container duration
            self.native_duration = self.container.duration
            self.time_base = Fraction(1, self.av.time_base)
            logger.debug("Using container instead of video stream duration")

        if not self.native_duration:
            raise BadFile(f"Cannot open {path}")

        # what to do with self.vstream.start_time?

        self.meta = {
            "width": vcc["width"],
            "height": vcc["height"],
            "duration": duration,  # container duration in seconds
            "fps": vcc["framerate"] or vcc["average_rate"],
            "sample_aspect_ratio": vcc["sample_aspect_ratio"] or Fraction(1, 1),
            "display_aspect_ratio": vcc["display_aspect_ratio"] or Fraction(vcc["width"], vcc["height"]),
            "format": vcc["pix_fmt"],
            # "container_bit_rate": self.container.bit_rate,
        }

    @staticmethod
    def import_backend():
        import av

        return av

    def iterall(self, native: bool = False) -> Iterator[Tuple[float, np.ndarray]]:
        for vframe in self.container.decode(self.vstream):  # can raise
            if native:
                frame = vframe
            else:
                frame = vframe.to_ndarray(format="rgb24")
            yield vframe.time, frame

    def _get_frame(self, offset: int, native: bool = False) -> Tuple[float, np.ndarray]:
        # if the stream duration could not be read, convert the offset from container time_base to stream time_base
        offset_in_corrected_time_base = int(offset * self.time_base / self.vstream.time_base)

        for _i in range(1):  # trying x times to find good frames
            try:
                self.container.seek(
                    offset_in_corrected_time_base, backward=True, any_frame=False, stream=self.vstream
                )  # this can silently fail for broken files
            except self.av.error.PermissionError:
                raise NoKeyFrame(f"Failed to seek to {offset} of {self.container.duration}.")

            try:
                for vframe in self.container.decode(self.vstream):  # can raise
                    if not vframe.is_corrupt:
                        logger.debug(
                            "Grabbing frame at %.03fs from seek offset %.03fs", vframe.time, offset * self.time_base
                        )

                        if native:
                            frame = vframe
                        else:
                            frame = vframe.to_ndarray(format="rgb24")

                        return vframe.time, frame

                    logger.debug(
                        "Skipping corrupt frame at %.03fs from seek offset %.03fs", vframe.time, offset * self.time_base
                    )

            except (self.av.error.InvalidDataError, self.av.error.ValueError) as e:
                logger.warning(
                    "Skipping corrupt container position at seek offset %.03fs: %s", offset * self.time_base, e
                )
                offset += int(self.time_base.denominator / self.meta["fps"])
            else:
                raise NoGoodFrame("Could not find good frame after decoding full file")

        raise NoGoodFrame("Could not find good frame after trying various offsets")

    def _frame_to_file(self, frame: "VideoFrame", outpath: Path) -> None:
        frame.to_image(fspath(outpath))

    def close(self) -> None:
        self.container.close()


def grab_pic(
    inpath: Union[str, PathLike],
    outpath: Path,
    pos: Union[float, Sequence[float]] = 0.5,
    overwrite: bool = False,
    backend: str = "cv",
) -> None:
    vf: VideoBase

    if backend == "av":
        vf = AvVideo(inpath)
    elif backend == "cv":
        vf = CvVideo(inpath)
    else:
        raise ValueError(f"Unsupported backend: {backend}")

    try:
        outpath.parent.mkdir(parents=True, exist_ok=True)

        if isinstance(pos, float):
            _raise_exists(outpath, overwrite)
            vf.save_frame_to_file(pos, outpath)
        elif isinstance(pos, list):
            for i, p in enumerate(pos):
                outpathseq = outpath.with_suffix(f".{i}{outpath.suffix}")
                _raise_exists(outpathseq, overwrite)
                vf.save_frame_to_file(p, outpathseq)
        else:
            raise TypeError("pos")

    finally:
        vf.close()


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("path")
    args = parser.parse_args()

    with CvVideo(args.path) as vf:
        print(vf.meta)

    with AvVideo(args.path) as vf:
        print(vf.meta)



from typing import Any, Iterable, Iterator, Optional, Sequence, TypeVar, Union

from typing_extensions import Self

T = TypeVar("T")


class _Default:
    pass


class BaseTask:
    def advance(self, delta: float) -> None:
        raise NotImplementedError

    def update(
        self,
        *,
        completed: Optional[float] = _Default,
        total: Optional[float] = _Default,
        description: Optional[str] = _Default,
        **fields: Any,
    ) -> None:
        raise NotImplementedError

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args):
        pass


class Task(BaseTask):
    def __init__(self, total: Optional[float] = None, description: Optional[str] = None) -> None:
        pass

    def advance(self, delta: float) -> None:
        pass

    def update(
        self,
        *,
        completed: Optional[float] = _Default,
        total: Optional[float] = _Default,
        description: Optional[str] = _Default,
        **fields: Any,
    ):
        pass


class Progress:
    def track(
        self,
        sequence: Union[Iterable[T], Sequence[T]],
        total: Optional[float] = None,
        description: Optional[str] = None,
        **fields: Any,
    ) -> Iterator[T]:
        return sequence

    def task(self, total: Optional[float] = None, description: Optional[str] = None, **fields: Any):
        return Task(total, description, **fields)

    def set_prolog(self, prolog: Any) -> None:
        pass

    def set_epilog(self, epilog: Any) -> None:
        pass

    def print(self, s: str, end="\n") -> None:
        print(s, end=end)



from typing import Any, Callable, Dict, Iterable, TypeVar, Union

import pyarrow as pa
from pyarrow import parquet as pq

T = TypeVar("T")


def table_dumps(table: pa.Table) -> bytes:
    writer = pa.BufferOutputStream()
    pq.write_table(table, writer)
    return bytes(writer.getvalue())


def table_loads(table: bytes) -> pa.Table:
    reader = pa.BufferReader(table)
    return pq.read_table(reader)


def schema_dumps(schema: pa.schema) -> bytes:
    writer = pa.BufferOutputStream()
    pq.write_metadata(schema, writer)
    return bytes(writer.getvalue())


pqmap = {
    "int64": pa.int64,
    "uint64": pa.uint64,
    "int32": pa.int32,
    "uint32": pa.uint32,
    "str": pa.string,
    "bool": pa.bool_,
}


def _to_pq_schema(d: Dict[str, Any], sort_keys: bool, outer: Callable[[list], T]) -> Union[T, pa.struct]:
    fields = []

    if sort_keys:
        keys: Iterable[str] = sorted(d)
    else:
        keys = d

    for k in keys:
        if isinstance(d[k], dict):
            fields.append((k, _to_pq_schema(d[k], sort_keys, pa.struct)))
        elif isinstance(d[k], list):
            if d[k]:
                if isinstance(d[k][0], dict):
                    fields.append((k, pa.list_(_to_pq_schema(d[k][0], sort_keys, pa.struct))))
                else:
                    fields.append((k, pa.list_(pqmap[d[k][0]]())))
        else:
            fields.append((k, pqmap[d[k]]()))

    return outer(fields)


def schema_simple_to_pq(schema: Dict[str, Any], sort_keys: bool = False) -> pa.schema:
    return _to_pq_schema(schema, sort_keys, pa.schema)



import logging
import mmap
import re
from struct import unpack
from typing import IO, Iterator, Optional, Set, Tuple, Union

from _hashlib import HASH as Hashobj

from ..exceptions import ParseError
from ..file import BufferedBinaryIoT, read_or_raise
from ..string import backslash_escaped_ascii

Segment = Union[Tuple[str, bytes], Tuple[str, bytes, bytes, bytes]]

segments = {
    # ISO/IEC 10918-1 : 1993(E), Table B.1
    # Start Of Frame markers, non-differential, Huffman coding
    b"\xff\xc0": ("SOF0", True, "Start of frame 0, baseline DCT"),
    b"\xff\xc1": ("SOF1", True, "Start of Frame 1, Extended Sequential DCT"),
    b"\xff\xc2": ("SOF2", True, "Start of frame 2, progressive DCT"),
    b"\xff\xc3": ("SOF3", True, "Start of Frame 3, Lossless (Sequential)"),
    # Start Of Frame markers, differential, Huffman coding
    b"\xff\xc5": ("SOF5", True, "Differential sequential DCT"),
    b"\xff\xc6": ("SOF6", True, "Differential progressive DCT"),
    b"\xff\xc7": ("SOF7", True, "Differential lossless (sequential)"),
    # Start Of Frame markers, non-differential, arithmetic coding"),
    b"\xff\xc8": ("JPG", True, "Reserved for JPEG extensions"),
    b"\xff\xc9": ("SOF9", True, "Extended sequential DCT"),
    b"\xff\xca": ("SOF10", True, "Progressive DCT"),
    b"\xff\xcb": ("SOF11", True, "Lossless (sequential)"),
    # Start Of Frame markers, differential, arithmetic coding
    b"\xff\xcd": ("SOF13", True, "Differential sequential DCT"),
    b"\xff\xce": ("SOF14", True, "Differential progressive DCT"),
    b"\xff\xcf": ("SOF15", True, "Differential lossless (sequential)"),
    # Huffman table specification
    b"\xff\xc4": ("DHT", True, "Define Huffman table(s)"),
    # Arithmetic coding conditioning specification
    b"\xff\xcc": ("DAC", True, "Define arithmetic coding conditioning(s)"),
    # Restart interval termination
    b"\xff\xd0": ("RST0", False, "Restart with modulo 8 count `m`"),
    b"\xff\xd1": ("RST1", False, "Restart with modulo 8 count `m`"),
    b"\xff\xd2": ("RST2", False, "Restart with modulo 8 count `m`"),
    b"\xff\xd3": ("RST3", False, "Restart with modulo 8 count `m`"),
    b"\xff\xd4": ("RST4", False, "Restart with modulo 8 count `m`"),
    b"\xff\xd5": ("RST5", False, "Restart with modulo 8 count `m`"),
    b"\xff\xd6": ("RST6", False, "Restart with modulo 8 count `m`"),
    b"\xff\xd7": ("RST7", False, "Restart with modulo 8 count `m`"),
    # Other markers
    b"\xff\xd8": ("SOI", False, "Start of image"),
    b"\xff\xd9": ("EOI", False, "End of image"),
    b"\xff\xda": ("SOS", True, "Start of scan"),
    b"\xff\xdb": ("DQT", True, "Define quantization table(s)"),
    b"\xff\xdc": ("DNL", True, "Define number of lines"),
    b"\xff\xdd": ("DRI", True, "Define restart interval"),
    b"\xff\xde": ("DHP", True, "Define hierarchical progression"),
    b"\xff\xdf": ("EXP", True, "Expand reference component(s)"),
    b"\xff\xe0": ("APP0", True, "Reserved for application segments"),
    b"\xff\xe1": ("APP1", True, "Reserved for application segments"),
    b"\xff\xe2": ("APP2", True, "Reserved for application segments"),
    b"\xff\xe3": ("APP3", True, "Reserved for application segments"),
    b"\xff\xe4": ("APP4", True, "Reserved for application segments"),
    b"\xff\xe5": ("APP5", True, "Reserved for application segments"),
    b"\xff\xe6": ("APP6", True, "Reserved for application segments"),
    b"\xff\xe7": ("APP7", True, "Reserved for application segments"),
    b"\xff\xe8": ("APP8", True, "Reserved for application segments"),
    b"\xff\xe9": ("APP9", True, "Reserved for application segments"),
    b"\xff\xea": ("APP10", True, "Reserved for application segments"),
    b"\xff\xeb": ("APP11", True, "Reserved for application segments"),
    b"\xff\xec": ("APP12", True, "Reserved for application segments"),
    b"\xff\xed": ("APP13", True, "Reserved for application segments"),
    b"\xff\xee": ("APP14", True, "Reserved for application segments"),
    b"\xff\xef": ("APP15", True, "Reserved for application segments"),
    b"\xff\xf0": ("JPG0", True, "Reserved for JPEG extensions"),
    b"\xff\xf1": ("JPG1", True, "Reserved for JPEG extensions"),
    b"\xff\xf2": ("JPG2", True, "Reserved for JPEG extensions"),
    b"\xff\xf3": ("JPG3", True, "Reserved for JPEG extensions"),
    b"\xff\xf4": ("JPG4", True, "Reserved for JPEG extensions"),
    b"\xff\xf5": ("JPG5", True, "Reserved for JPEG extensions"),
    b"\xff\xf6": ("JPG6", True, "Reserved for JPEG extensions"),
    b"\xff\xf7": ("JPG7", True, "Reserved for JPEG extensions"),
    b"\xff\xf8": ("JPG8", True, "Reserved for JPEG extensions"),
    b"\xff\xf9": ("JPG9", True, "Reserved for JPEG extensions"),
    b"\xff\xfa": ("JPG10", True, "Reserved for JPEG extensions"),
    b"\xff\xfb": ("JPG11", True, "Reserved for JPEG extensions"),
    b"\xff\xfc": ("JPG12", True, "Reserved for JPEG extensions"),
    b"\xff\xfd": ("JPG13", True, "Reserved for JPEG extensions"),
    b"\xff\xfe": ("COM", True, "Comment"),
    # Reserved markers
    b"\xff\x01": ("TEM", False, "For temporary private use in arithmetic coding"),
    # b"\xFF\x02" to b"\xFF\xBF": ("RES", True, "Reserved"),
}

markerp = re.compile(b"\xff[^\x00]")  # don't use raw literals

metadata_segments = {
    "COM",
    "TRAILER",
} | {f"APP{i}" for i in range(16)}


def iter_jpeg_fp(fr: IO[bytes], translate: bool = True) -> Iterator[Segment]:
    """Iterate over JPEG file given in binary stream `fr` and yield the segments.
    `translate=False` can be used to reconstruct the file in a bit-identical way.
    See copy_jpeg() for an example.
    """

    with mmap.mmap(fr.fileno(), 0, access=mmap.ACCESS_READ) as mm:
        fs = mm.size()

        eoi = False

        while True:
            if not eoi:
                marker = read_or_raise(mm, 2)
            else:
                data = mm.read()
                if data:
                    if translate:
                        yield "TRAILER", data
                    else:
                        yield "TRAILER", b"", b"", data
                return

            try:
                info = segments[marker]
            except KeyError:
                raise ParseError(f"Invalid segment marker: {backslash_escaped_ascii(marker)} at {mm.tell()}")

            name = info[0]
            hasdata = info[1]

            if name == "SOS":
                size = read_or_raise(mm, 2)
                (length,) = unpack(">H", size)
                data = read_or_raise(mm, length - 2)

                if translate:
                    yield name, data
                else:
                    yield name, marker, size, data

                prevenv = mm.tell()
                for m in markerp.finditer(mm, prevenv, fs):
                    start, end = m.span(0)
                    if translate:
                        yield "ENTROPY", mm[prevenv:start]
                    else:
                        yield "ENTROPY", b"", b"", mm[prevenv:start]
                    prevenv = end

                    marker = m.group(0)

                    try:
                        info = segments[marker]
                    except KeyError:
                        raise ParseError(f"Invalid segment marker: {backslash_escaped_ascii(marker)} at {start}")

                    name = info[0]

                    if name[:3] == "RST":
                        if translate:
                            yield name, b""
                        else:
                            yield name, marker, b"", b""
                    else:
                        mm.seek(start)  # this is seeking back
                        break

            else:
                eoi = eoi or name == "EOI"

                if hasdata:
                    size = read_or_raise(mm, 2)
                    (length,) = unpack(">H", size)
                    data = read_or_raise(mm, length - 2)
                else:
                    size = b""
                    data = b""

                if translate:
                    yield name, data
                else:
                    yield name, marker, size, data


def iter_jpeg(path: str, translate: bool = True) -> Iterator[Segment]:
    """Same as `iter_jpeg_fp()` except that it accepts a path."""

    with open(path, "rb") as fr:
        yield from iter_jpeg_fp(fr, translate=translate)


def copy_jpeg_fp(fin: IO[bytes], fout: BufferedBinaryIoT, ignore_segments: Optional[Set[str]] = None) -> None:
    """Same as `copy_jpeg()` except that it accepts file-like objects."""

    ignore_segments = ignore_segments or set()

    for name, marker, size, data in iter_jpeg_fp(fin, translate=False):
        if name not in ignore_segments:
            fout.write(marker)
            fout.write(size)
            fout.write(data)


def copy_jpeg(inpath: str, outpath: str, ignore_segments: Optional[Set[str]] = None) -> None:
    """Copy JPEG file `inpath` to `outpath` while ignoring the JPEG segments
    given in `ignore_segments`.

    Example:
            copy_jpeg("in.jpg", "out.jpg", {"COM", "TRAILER"})
    """

    with open(inpath, "rb") as fr, open(outpath, "xb") as fw:
        copy_jpeg_fp(fr, fw, ignore_segments=ignore_segments)


def hash_raw_jpeg(path: str, hashobj: Hashobj) -> None:
    """Create a hash of the JPEG at `path` skipping over meta data sections."""

    for name, marker, size, data in iter_jpeg(path, translate=False):
        if name not in metadata_segments:
            hashobj.update(marker)
            hashobj.update(size)
            hashobj.update(data)


if __name__ == "__main__":
    from argparse import ArgumentParser
    from hashlib import sha1

    from genutility.args import is_dir
    from genutility.hash import hashsum_file_format
    from genutility.iter import consume

    parser = ArgumentParser()
    parser.add_argument("path", type=is_dir)
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-r", "--recursive", action="store_true")
    parser.add_argument("--hash", action="store_true")
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    if args.recursive:
        it = args.path.rglob("*.jpg")
    else:
        it = args.path.glob("*.jpg")

    valid = 0
    invalid = 0

    for path in it:
        try:
            if args.hash:
                hashobj = sha1()  # nosec
                hash_raw_jpeg(path, hashobj)
                print(hashsum_file_format(hashobj, path))
            else:
                consume(iter_jpeg(path))

        except ParseError as e:
            logging.debug("Parse error in %s: %s", path, e)
            invalid += 1

        except EOFError:
            logging.debug("Truncated file %s", path)
            invalid += 1

        else:
            valid += 1

    print("valid", valid, "invalid", invalid)



from random import choice, random, randrange, sample
from typing import Iterator, Sequence, Tuple


def randstr(length: int, charset: str) -> str:
    """Returns a (noncryptographic) random string consisting of characters from `charset`
    of length `length`.
    """

    return "".join(choice(charset) for i in range(length))  # nosec


def randbytes(size: int) -> bytes:
    """Returns (noncryptographic) random bytes of length `length`."""

    return bytes(randrange(0, 256) for _ in range(size))  # nosec


def rgb_colors() -> Iterator[Tuple[int, int, int]]:
    """Yields a stream of (noncryptographic) random RGB color tuples."""

    while True:
        rgb = randrange(0, 256**3)  # nosec
        rg, b = divmod(rgb, 256)
        r, g = divmod(rg, 256)
        yield (r, g, b)


def randomized(seq: Sequence) -> Sequence:
    """Like `random.shuffle`, but not in-place."""

    return sample(seq, len(seq))


def prob_false(probability: float) -> bool:
    """Returns `False` with a probability of `probability`."""

    if probability in (0, 1):
        return not bool(probability)  # for weird open interval edge cases

    return probability < random()  # random() -> [0.0, 1.0) # nosec



import os.path
import re
from string import ascii_letters, digits
from typing import Dict, Iterable, Optional
from urllib.parse import SplitResult, parse_qs, urlencode, urlsplit, urlunsplit

# URI RFC
gen_delims = ":/?#[]@"
sub_delims = "!$&'()*+;="
reserved = gen_delims + sub_delims
unreserved = ascii_letters + digits + "-._~"
valid_uri_characters = reserved + unreserved

uri_schemes = ("http", "https", "ftp", "sftp", "irc", "magnet", "file", "data")
url_base_pattern = r"(?:(?:{}:\/\/)|www\.)(?:[{}]|%[a-zA-Z0-9]{{2}})+"


def get_url_pattern(schemes: Optional[Iterable[str]] = None) -> re.Pattern:
    schemes = schemes or uri_schemes
    return re.compile(url_base_pattern.format("|".join(schemes), re.escape(valid_uri_characters)))


def get_filename_from_url(url: str, strip: Optional[str] = None) -> str:
    return urlsplit(url).path.rstrip(strip).rsplit("/", 1)[1]


def get_url_argument(split_url: SplitResult, argument: str, path: Optional[str] = None) -> Optional[str]:
    """Extracts a parameter value from the url query string. Optionally compares the url path.
    For example:
    > get_url_argument(urlsplit("https://www.google.com/search?q=python"), "q")
    'python'
    """

    if not path or split_url.path == path:
        try:
            return parse_qs(split_url.query)[argument][0]
        except KeyError:
            return None
    return None


def url_replace_query(split_url: SplitResult, query: Dict[str, str], drop_fragment: bool = True) -> str:
    """Replace the query part of an URL with a new one.
    For example:
    > url_replace_query(urlsplit("https://www.google.com/search?source=hp&q=java"), {"q": "python"})
    'https://www.google.com/search?q=python'
    """

    scheme, netloc, path, _, fragment = split_url

    _query = urlencode(query)
    if drop_fragment:
        fragment = ""

    return urlunsplit((scheme, netloc, path, _query, fragment))


def path_ext(path: str) -> str:
    return os.path.splitext(path)[1][1:].lower()


def url_ext(url: str) -> str:
    """Returns the lowercase file extension of the URI/URL."""

    path = urlsplit(url).path
    return path_ext(path)



from base64 import urlsafe_b64decode, urlsafe_b64encode
from typing import Dict, Hashable, Optional, TypeVar
from urllib.parse import quote, unquote

from flask import Response, request
from werkzeug.routing import BaseConverter

T = TypeVar("T")
H = TypeVar("H", bound=Hashable)

""" examples

user_dict = {
    "username": "password"
}

@before_request
def auth():
    return do_basic_auth(user_dict)

"""


class Base64Converter(BaseConverter):
    def to_python(self, value: str) -> str:  # decode once as the wsgi app already receives a decoded url
        return urlsafe_b64decode(value.encode("ascii")).decode("utf-8")

    def to_url(self, value: str) -> str:
        return urlsafe_b64encode(value.encode("utf-8")).decode("ascii")


class DoubleQuoteConverter(BaseConverter):
    def to_python(self, value: str) -> str:  # decode once as the wsgi app already receives a decoded url
        return unquote(value)

    def to_url(self, value: str) -> str:
        return quote(quote(value, safe=""), safe="")


def check_auth(username: H, password: T, user_dict: Dict[H, T]) -> bool:
    try:
        return user_dict[username] == password
    except KeyError:
        return False


def authenticate() -> Response:
    """Sends a 401 response that enables basic auth"""

    return Response(
        "Could not verify your access level for that URL.\n" "You have to login with proper credentials",
        401,
        {"WWW-Authenticate": 'Basic realm="Login Required"'},
    )


def do_basic_auth(user_dict: Dict[H, T]) -> Optional[Response]:
    auth = request.authorization
    if not auth or not check_auth(auth.username, auth.password, user_dict):
        return authenticate()
    else:
        return None



from typing import Union

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.dh import DHPrivateKey
from cryptography.hazmat.primitives.asymmetric.dsa import DSAPrivateKey
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, generate_private_key
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
    load_der_private_key,
    load_pem_private_key,
)
from OpenSSL import crypto  # pyOpenSSL
from OpenSSL.crypto import X509

from .exceptions import assert_choice

PrivateKey = Union[RSAPrivateKey, DSAPrivateKey, DHPrivateKey, EllipticCurvePrivateKey]


def load_certificate_file(filename: str) -> X509:
    with open(filename, "rb") as fr:
        return crypto.load_certificate(crypto.FILETYPE_PEM, fr.read())


def get_pubkey_from_x509(x509: X509) -> bytes:
    return crypto.dump_privatekey(crypto.FILETYPE_ASN1, x509.get_pubkey())


def load_keyfile(filename: str, encoding: str = "PEM") -> PrivateKey:
    try:
        load_func = {
            "PEM": load_pem_private_key,
            "DER": load_der_private_key,
        }[encoding]
    except KeyError:
        raise ValueError("Invalid encoding")

    with open(filename, "rb") as fr:
        password = None
        backend = default_backend()
        return load_func(fr.read(), password, backend)


def generate_rsa_keyfile_pair(
    priv_key: str, pub_key: str, key_size: int = 4096, encoding: str = "PEM", format: str = "modern"
) -> RSAPrivateKey:
    """
    encoding: "PEM" -> ASN.1 encoding, "DER" -> base64 PEM
    """

    if key_size < 2048:
        raise ValueError("key_size must be atleast 2048")

    encodings = {
        "PEM": Encoding.PEM,
        "DER": Encoding.DER,
    }

    private_formats = {
        "modern": PrivateFormat.PKCS8,
        "legacy": PrivateFormat.TraditionalOpenSSL,
    }

    public_formats = {
        "modern": PublicFormat.SubjectPublicKeyInfo,
        "legacy": PublicFormat.PKCS1,
    }

    assert_choice("encoding", encoding, encodings)
    assert_choice("format", format, private_formats)

    public_exponent = 65537
    backend = default_backend()

    key = generate_private_key(public_exponent, key_size, backend)

    with open(priv_key, "wb") as fw:
        fw.write(key.private_bytes(encodings[encoding], private_formats[format], NoEncryption()))

    with open(pub_key, "wb") as fw:
        fw.write(key.public_key().public_bytes(encodings[encoding], public_formats[format]))

    return key



from math import sqrt

import numpy as np

from .numba import opjit
from .numpy import is_square


@opjit(["void(f4[:,:], f4[:], f4)", "void(f8[:,:], f8[:], f8)"])
def _choldate(L, x, sign):
    for k in range(L.shape[0]):
        r = sqrt(L[k, k] * L[k, k] + sign * x[k] * x[k])
        c = r / L[k, k]
        s = x[k] / L[k, k]
        L[k, k] = r

        L[k + 1 :, k] = (L[k + 1 :, k] + sign * s * x[k + 1 :]) / c
        x[k + 1 :] = c * x[k + 1 :] - s * L[k + 1 :, k]


def choldate(L: np.ndarray, x: np.ndarray, sign: str) -> None:
    """This function computes the lower triangular Cholesky decomposition L' of matrix A' from L in-place
    (the cholesky decomp of A) where A' = A + sign*x*x^T.
    Based on: https://en.wikipedia.org/wiki/Cholesky_decomposition#Rank-one_update

    L: square lower triangular matrix
    """

    if not is_square(L) or x.shape != (L.shape[0],):
        raise ValueError("Invalid dimensions")

    try:
        _sign = {"+": +1.0, "-": -1.0}[sign]
    except KeyError:
        raise ValueError("Invalid sign")

    return _choldate(L, x, _sign)


if __name__ == "__main__":
    import timeit

    from .numpy import random_triangular_matrix

    size = 100
    L = random_triangular_matrix(size)
    x = np.random.uniform(0, 1, size)

    choldate(L, x, "+")  # warmup
    print(min(timeit.repeat('choldate(L, x, "+")', number=10000, repeat=5, globals=globals())))



from os import remove
from tempfile import NamedTemporaryFile
from typing import Optional

from .file import _check_arguments


class CloseableNamedTemporaryFile:
    def __init__(self, mode: str = "w+b", encoding: Optional[str] = None) -> None:
        encoding = _check_arguments(mode, encoding)
        self.f = NamedTemporaryFile(mode=mode, encoding=encoding, delete=False)

    def __enter__(self):
        self.f.__enter__()
        return self.f, self.f.name

    def __exit__(self, exc_type, exc_value, traceback):
        self.f.__exit__(exc_type, exc_value, traceback)
        remove(self.f.name)



from operator import itemgetter
from typing import Any, Dict, Optional

import pandas as pd
import requests

from .dict import mapmap


def wikidata_to_dataframe(result: Dict[str, Any]) -> pd.DataFrame:
    cols = result["head"]["vars"]

    getval = itemgetter("value")

    def gen():
        for row in result["results"]["bindings"]:
            yield tuple(map(getval, mapmap(row, cols)))

    return pd.DataFrame.from_records(gen(), columns=cols)


def query_wikidata(query: str, timeout: Optional[float] = 120.0) -> Dict[str, Any]:
    url = "https://query.wikidata.org/sparql"
    r = requests.get(url, params={"format": "json", "query": query}, timeout=timeout)
    r.raise_for_status()
    return r.json()



class SplitRecursive:
    """Fast non-prime-based factorial algorithm.
    Python port of: http://www.luschny.de/math/factorial/java/FactorialSplit.java.html
    """

    def factorial(self, n: int) -> int:
        if n < 0:
            raise ArithmeticError(f"Factorial: n has to be >= 0, but was {n}")

        if n < 2:
            return 1

        p = 1
        r = 1
        self.N = 1

        log2n = n.bit_length() - 1  # should be the same as: int(floor(log(n, 2)))
        h = 0
        shift = 0
        high = 1

        while h != n:
            shift += h
            h = n >> log2n
            log2n -= 1
            length = high
            if (h & 1) == 1:
                high = h
            else:
                high = h - 1
            length = (high - length) >> 1

            if length > 0:
                p *= self.product(length)
                r *= p
        return r << shift

    def product(self, n: int) -> int:
        m = n >> 1
        if m == 0:
            self.N += 2
            return self.N
        if n == 2:
            self.N += 2
            ret = self.N * (self.N + 2)
            self.N += 2
            return ret
        return self.product(n - m) * self.product(m)



from collections import UserDict, defaultdict
from collections.abc import Mapping, MutableMapping  # noqa: F401
from copy import deepcopy
from typing import Any, Callable, Dict, Hashable, Iterable, Iterator, List
from typing import Mapping as MappingT
from typing import MutableMapping as MutableMappingT
from typing import Tuple, TypeVar, Union

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")


def flatten(d: Union[Dict[T, U], List[U], Tuple[U, ...]]) -> Iterator[U]:
    """Flattens dicts/lists of (dicts/lists of ...) lists to lists."""

    if isinstance(d, (list, tuple)):
        yield from d
    elif isinstance(d, dict):
        for val in d.values():
            yield from flatten(val)
    else:
        raise TypeError(f"Unsupported type: {type(d)}")


def _flatten_keys(d: Dict[Any, Any], out: Dict[tuple, Any], path: tuple) -> None:
    for k, v in d.items():
        if isinstance(v, dict):
            _flatten_keys(v, out, path + (k,))
        else:
            out[path + (k,)] = v


def flatten_keys(d: Dict[Any, Any]) -> Dict[tuple, Any]:
    out: Dict[tuple, Any] = {}
    _flatten_keys(d, out, ())
    return out


def rec_update(d: MutableMappingT, u: MappingT) -> None:
    """Updates dict `d` recursively with values from `u`.
    Requires `setdefault` method.
    """

    if not isinstance(d, MutableMapping) or not isinstance(u, Mapping):
        raise TypeError("All arguments must be mappings")

    for k, v in u.items():
        if isinstance(v, Mapping):
            rec_update(d.setdefault(k, {}), v)
        else:
            d[k] = v


def rec_update_mod(d: Any, u: MappingT) -> None:
    """Similar to `rec_update`, except that `d` doesn't require the `setdefault` method."""

    for k, v in u.items():
        if isinstance(v, Mapping):
            try:
                sub = d[k]
            except KeyError:
                # d might not store a reference to the original object, thus don't do `sub = d[k] = {}`
                d[k] = {}
                sub = d[k]
            rec_update_mod(sub, v)
        else:
            d[k] = v


def get_one_of(d: Dict[T, U], keys: Iterable[T]) -> Tuple[T, U]:
    """Returns the (key, value) pair of the first key of `keys` found in `d`."""

    for key in keys:
        try:
            return key, d[key]
        except KeyError:
            pass

    raise KeyError("None of the keys could be found")


# similar: subdict
def get_available(d: Dict[T, U], keys: Iterable[T]) -> Iterator[Tuple[T, U]]:
    """Returns all the key-value pairs in `d` for the keys in `it`.
    Missing keys are ignored.
    """

    for key in keys:
        try:
            yield key, d[key]
        except KeyError:
            pass


def subdict(d: MappingT[T, U], it: Iterable[T]) -> Dict[T, U]:
    """Uses the elements of `it` as keys to extract a new sub-dictionary.
    Raises if not all keys in `it` are available.
    """

    return {key: d[key] for key in it}


# was: mapdict, mapget
def mapmap(d: MappingT[T, U], it: Iterable[T]) -> Iterator[U]:
    """Returns all the values of `d` for the keys in `it`.
    Raises for missing keys.
    """

    return (d[i] for i in it)


def hasvalues(d: dict) -> dict:
    """Returns a sub-dictionary which leaves out all pairs where the value evaluates to False."""

    return {k: v for k, v in d.items() if v}


def valuemap(func: Callable[[U], V], d: Dict[T, U]) -> Dict[T, V]:
    """Returns a new dictionary with `func` applied to all values of `d`."""

    return {k: func(v) for k, v in d.items()}


def itemgetter(it: Iterable[T]) -> Callable[[MappingT[T, U]], Iterator[U]]:
    """Similar to `operator.itemgetter` except that it always expects and returns iterables.
    Compare `mapmap`
    """

    return lambda d: (d[i] for i in it)


def subdictdefault(d: MappingT[T, U], it: Iterable[T], default: V = None) -> Dict[T, Union[U, V]]:
    """Uses the elements of `it` as keys to extract a new sub-dictionary."""

    return {key: d.get(key, default) for key in it}


def update(d1: dict, d2: dict) -> None:
    """Same as `dict.update` except that `None` values are skipped."""

    for k, v in d2.items():
        if v is not None:
            d1[k] = v


class keydefaultdict(defaultdict):
    """defaultdict which passes a key to the default factory."""

    def __missing__(self, key: Hashable) -> Any:
        if self.default_factory is None:
            raise KeyError(key)
        else:
            value = self[key] = self.default_factory(key)  # type: ignore[call-arg]
            return value


class KeyExistsError(KeyError):
    pass


class NoOverwriteDict(UserDict):
    """Dictionary which does not allow overwriting existing items."""

    def __setitem__(self, key: Hashable, value: Any) -> None:
        if key in self.data:
            raise KeyExistsError(repr(key))
        self.data[key] = value

    def overwrite(self, key: Hashable, value: Any) -> None:
        self.data[key] = value

    def dict(self):
        return self.data


def _merge_schema(d1: dict, d2: dict) -> None:
    """Merge `d2` into `d1`. `d2` stays unmodified."""

    a = d1.keys()
    b = d2.keys()

    for k in a & b:
        if type(d1[k]) != type(d2[k]):  # noqa: E721
            raise TypeError(f"Type of `{k}` changed from {type(d1[k])} to {type(d2[k])}")

        if isinstance(d1[k], list):
            if d1[k] and d2[k]:
                if type(d1[k][0]) != type(d2[k][0]):  # noqa: E721
                    raise TypeError(f"Type of list `{k}` changed from {type(d1[k][0])} to {type(d2[k][0])}")
                if isinstance(d1[k][0], dict):
                    _merge_schema(d1[k][0], d2[k][0])

        elif isinstance(d1[k], dict):
            _merge_schema(d1[k], d2[k])

        elif isinstance(d1[k], int):
            d1[k] = max(d1[k], d2[k])

    for k in b - a:
        if isinstance(d2[k], list):
            d1[k] = deepcopy(d2[k][:1])
        else:
            d1[k] = deepcopy(d2[k])


def _get_intsize(num: int) -> str:
    if num >= 2**63:
        return "uint64"
    elif num >= 2**32:
        return "int64"
    elif num >= 2**31:
        return "uint32"
    else:
        return "int32"


def _post_schema(d: dict) -> None:
    for k in d:
        if isinstance(d[k], dict):
            _post_schema(d[k])
        elif isinstance(d[k], list):
            if d[k]:
                if isinstance(d[k][0], dict):
                    _post_schema(d[k][0])
                elif isinstance(d[k][0], int):
                    d[k][0] = _get_intsize(d[k][0])
                else:
                    d[k][0] = type(d[k][0]).__name__
        elif isinstance(d[k], int):
            d[k] = _get_intsize(d[k])
        else:
            d[k] = type(d[k]).__name__


def get_schema_simple(d: Iterable[dict]) -> dict:
    """Returns a combined schema definition for the dicts provided by iterable `d`.
    - Keys are assumed to be optional.
    - Type unions are not supported. Changing field types will throw an exception.
    - Ints are combined to the largest type (eg. int32->int64)
    - For lists, only the first element is looked at, and it is assumed that all following elements
        are of the same type. This is not verified however.
    """

    schema: Dict[str, Any] = {}
    for i in d:
        _merge_schema(schema, i)

    _post_schema(schema)
    return schema


# was: zipmap
def zipget(objs: Iterable[MappingT[T, U]], keys: Iterable[T]) -> Iterator[U]:
    """Gets a list of keys from a list of mappings.
        zipget([[1, 2], [3, 4]], [0, 1]) -> (1, 4)

    Similar to numpy indexing:
        a = np.array([[1, 2], [3, 4]])
        a[np.arange(len(a)), [0, 1]] -> array([1, 4])
    """

    return (obj[key] for obj, key in zip(objs, keys))



from collections import defaultdict
from pathlib import Path
from sys import maxunicode
from tempfile import gettempdir
from unicodedata import category

from .cache import cache  # nosec


@cache(Path(gettempdir()) / "unicode_categories.{protocol}.pkl", ignoreargs=True)
def unicode_categories():
    unicode_categories = defaultdict(set)
    for c in map(chr, range(maxunicode + 1)):
        unicode_categories[category(c)].add(c)
    return unicode_categories



from .mp4 import enumerate_atoms


class NoExifFound(Exception):
    pass


def heif_get_exif(path: str) -> bytes:
    item_id = None
    iloc = None
    entries = None

    for _depth, _pos, type, _size, content, _leaf in enumerate_atoms(path, parse_atoms=True, unparsed_data=True):
        if type == "infe":
            if content["item_type"] == b"Exif":
                item_id = content["item_ID"]

        elif type == "iloc":
            iloc = content

    if not item_id or not iloc:
        raise NoExifFound("No Exif info found")

    for item in iloc["items"]:
        if item["item_ID"] == item_id:
            entries = item["entries"]

    if not entries or len(entries) != 1:
        raise ValueError("Bad file")

    extent_offset = entries[0].extent_offset + 4
    extent_length = entries[0].extent_length

    with open(path, "rb") as fr:
        fr.seek(extent_offset)
        return fr.read(extent_length)  # buffered


if __name__ == "__main__":
    import logging
    from argparse import ArgumentParser

    import piexif

    from genutility.args import is_dir

    parser = ArgumentParser()
    parser.add_argument("path", type=is_dir)
    parser.add_argument("-r", "--recursive", action="store_true")
    args = parser.parse_args()

    if args.recursive:
        it = args.path.rglob("*.heic")
    else:
        it = args.path.glob("*.heic")

    for path in it:
        relpath = path.relative_to(args.path)

        try:
            data = heif_get_exif(path)
        except NoExifFound:
            logging.info("%s: No Exif Found", relpath)
        except ValueError:
            logging.exception(relpath)
        except AssertionError as e:
            logging.error("%s: AssertionError: %s", relpath, e)
        else:
            exif = piexif.load(data)
            print(
                relpath,
                exif["0th"][271].decode("ascii"),
                exif["0th"][272].decode("ascii"),
                exif["0th"][piexif.ImageIFD.DateTime].decode("ascii"),
            )



import json
import ssl
from typing import Callable, List, Optional, Tuple

from .http import URLRequest
from .ops import logical_xor


class StreamWatcher:
    def __init__(self, api: "TwitchAPI") -> None:
        self.api = api
        self.followed_names = api.get_followed()
        self.followed_online = {userid: False for userid in self.followed_names.keys()}

    def watch(
        self, notify_started: Callable[[str, str, Optional[str]], None], notify_stopped: Callable[[str, str], None]
    ) -> None:
        user_ids = self.followed_names.keys()
        streams = self.api.get_streams(user_ids)
        online = streams.keys()
        offline = user_ids - online

        for user_id in online:
            if not self.followed_online[user_id]:
                self.followed_online[user_id] = True
                notify_started(user_id, self.followed_names[user_id], streams[user_id].get("title", None))

        for user_id in offline:
            if self.followed_online[user_id]:
                self.followed_online[user_id] = False
                notify_stopped(user_id, self.followed_names[user_id])


class TwitchAPI:
    base = "https://api.twitch.tv/helix/"
    login = "users"
    follows = "users/follows"
    streams = "streams?user_id={}"

    def __init__(
        self,
        client_id: str,
        userid: Optional[str] = None,
        username: Optional[str] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
    ) -> None:
        if not logical_xor(userid, username):
            raise ValueError("Either the userid or the username must be given")

        self.client_id = client_id
        self.ssl_context = ssl_context

        if userid:
            self.userid = userid
        else:
            assert username
            self.userid = self.get_userid(username)

    def req(self, url: str, params: List[Tuple[str, str]]) -> dict:
        qs = "&".join(k + "=" + v for k, v in params)
        data = URLRequest(url + "?" + qs, headers={"Client-ID": self.client_id}, context=self.ssl_context).load()
        return json.loads(data)

    def get_userid(self, username: str) -> str:
        d = self.req(self.base + self.login, [("login", username)])
        return d["data"][0]["id"]

    def get_followed(self):
        d = self.req(self.base + self.follows, [("from_id", self.userid)])
        return {follow["to_id"]: follow["to_name"] for follow in d["data"]}

    def get_streams(self, user_ids):
        d = self.req(self.base + self.streams, [("user_id", user_id) for user_id in user_ids])
        ret = {stream["user_id"]: stream for stream in d["data"]}
        assert len(d["data"]) == len(ret), "More than one stream per user"
        return ret

    def watcher(self) -> StreamWatcher:
        return StreamWatcher(self)



from pymongo.collection import Collection

from .exceptions import InconsistentState, NoResult


def findone(db: Collection, query: dict) -> dict:
    docs = list(db.find(query))

    if len(docs) == 0:
        raise NoResult("No result found")
    elif len(docs) == 1:
        return docs[0]
    else:
        raise InconsistentState("More than one result found")



from typing import Any

import toml


def read_toml(path: str) -> Any:
    with open(path, encoding="utf-8") as fr:
        return toml.load(fr)


def write_toml(obj: Any, path: str) -> None:
    with open(path, "w", encoding="utf-8") as fw:
        toml.dump(obj, fw)



from os import PathLike, fspath
from pathlib import Path
from typing import Any, Dict, Union

from pymediainfo import MediaInfo, Track


class MediaInfoHelper:
    def __init__(self, path: Union[str, Path]) -> None:
        if isinstance(path, PathLike):
            path = fspath(path)

        self.mi = MediaInfo.parse(path)

    def audio_duration(self) -> float:
        """Returns audio duration in seconds."""

        for track in self.mi.tracks:
            if track.track_type == "Audio":
                return track.duration / 1000.0

        raise LookupError("Audio duration not available")

    @property
    def general(self):
        return self.first_track("General")

    @property
    def video(self):
        return self.first_track("Video")

    @property
    def audio(self):
        return self.first_track("Audio")

    @property
    def text(self):
        return self.first_track("Text")

    def first_track(self, track_type: str) -> Track:
        for track in self.mi.tracks:
            if track.track_type == track_type:
                return track

        raise LookupError("Video track not available")

    def is_vfr(self) -> bool:
        for track in self.mi.tracks:
            if track.track_type == "Video":
                return track.frame_rate_mode == "VFR"

        return False

    def meta_info(self) -> Dict[str, Any]:
        for track in self.mi.tracks:
            if track.track_type == "General":
                if track.duration is None:
                    duration = None
                else:
                    duration = track.duration / 1000.0

                return {
                    "title": track.title,  # track_name
                    "performer": track.performer,
                    "album": track.album,
                    "track-position": track.track_name_position,
                    "duration": duration,
                    "date": track.Recorded_Date,
                }

        raise LookupError("Meta data not available")



import bs4


def simple_microdata_parser(bstree: bs4.BeautifulSoup) -> dict:  # schema.org Microdata, replace with `extruct` module?
    def is_item_scope(elem):
        return elem.has_attr("itemscope") and elem.has_attr("itemtype") and not elem.has_attr("itemprop")

    def is_item_scopeprop(elem):
        return elem.has_attr("itemscope") and elem.has_attr("itemtype") and elem.has_attr("itemprop")

    def is_item_prop(elem):
        return elem.has_attr("itemprop") and not elem.has_attr("itemscope") and not elem.has_attr("itemtype")

    def rec(tree, dict=None):
        if dict is None:
            dict = {}

        for child in tree.children:
            if isinstance(child, bs4.element.Tag):
                if is_item_scope(child):
                    key = child["itemtype"]
                    value = rec(child)
                    # print("tag {} is scope: {}".format(child.name, child["itemtype"]))

                    try:
                        dict[key].append(value)
                    except KeyError:
                        dict[key] = [value]

                elif is_item_scopeprop(child):
                    # print("tag {} is scopeprop: {}, {}".format(child.name, child["itemtype"], child["itemprop"]))

                    try:
                        dict[(child["itemtype"], child["itemprop"])].append(rec(child))
                    except KeyError:
                        dict[(child["itemtype"], child["itemprop"])] = [rec(child)]

                elif is_item_prop(child):
                    key = child["itemprop"]

                    if child.name == "meta":
                        value = child["content"]
                    elif key == "url":
                        value = child["href"]
                        rec(child, dict)
                    elif child.contents:
                        value = child.text.strip()  # was: child.contents[0].string
                    else:
                        continue

                    try:
                        dict[key].append(value)
                    except KeyError:
                        dict[key] = [value]

                else:
                    rec(child, dict)

        return dict

    return rec(bstree)  # .html



import asyncio
import csv
import logging
import re
import time
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Coroutine, Dict, Generic, Iterable, Iterator, List
from typing import Mapping as MappingT
from typing import Optional
from typing import Sequence as SequenceT
from typing import Tuple, TypeVar, Union

import aiohttp
import requests
from simple_salesforce import Salesforce
from simple_salesforce.exceptions import SalesforceExpiredSession

from .atomic import sopen
from .callbacks import Progress
from .json import read_json, write_json

if TYPE_CHECKING:
    import pandas as pd  # noqa: F401

JsonDict = Dict[str, Any]
ReturnTGet = TypeVar("ReturnTGet")
ReturnTPost = TypeVar("ReturnTPost")
T = TypeVar("T")

logger = logging.getLogger(__name__)

_sosl_pat = re.compile("[" + re.escape("?&|!{}[]()^~*:\\\"'+-") + "]")


def _sosl_repl(m) -> str:
    return "\\" + m.group(0)


def sosl_escape(s: str) -> str:
    return _sosl_pat.sub(_sosl_repl, s)


def one(result: SequenceT[T]) -> T:
    if len(result) == 1:
        return result[0]
    else:
        raise ValueError("More than one result")


class SalesforceError(Exception):
    pass


def _flatten_row(row):
    ret = {}
    del row["attributes"]

    for k, v in row.items():
        if isinstance(v, dict):
            for kk, vv in _flatten_row(v).items():
                ret[f"{k}.{kk}"] = vv
        else:
            ret[k] = v

    return ret


class MySalesforce:
    def __init__(
        self,
        username: str,
        password: str,
        security_token: str,
        consumer_key: str,
        consumer_secret: str,
        test: bool = False,
        cache_file: Optional[str] = None,
        timeout: int = 60,
    ) -> None:
        self.username = username
        self.password = password
        self.security_token = security_token
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.test = test
        self.cache_file = cache_file
        self.timeout = timeout

        self._session: Optional[Salesforce] = None

    def _login(self) -> Tuple[str, str]:
        logger.debug("Obtaining new Salesforce credentials")

        if self.test:
            url = "https://test.salesforce.com/services/oauth2/token"
        else:
            url = "https://login.salesforce.com/services/oauth2/token"

        params = {
            "grant_type": "password",
            "client_id": self.consumer_key,
            "client_secret": self.consumer_secret,
            "username": self.username,
            "password": self.password + self.security_token,
        }

        r = requests.post(url, params=params, timeout=self.timeout)
        r.raise_for_status()
        response = r.json()

        return response["instance_url"], response["access_token"]

    def session(self, fresh: bool = False) -> Salesforce:
        if self._session is not None and not fresh:
            return self._session

        try:
            if fresh or not self.cache_file:
                raise FileNotFoundError

            obj = read_json(self.cache_file)
            logger.debug("Creating new Salesforce session")
            self._session = Salesforce(instance_url=obj["instance_url"], session_id=obj["session_id"])

        except FileNotFoundError:
            instance_url, session_id = self._login()

            if self.cache_file:
                write_json(
                    {
                        "instance_url": instance_url,
                        "session_id": session_id,
                    },
                    self.cache_file,
                )

            logger.debug("Creating new Salesforce session")
            self._session = Salesforce(instance_url=instance_url, session_id=session_id)

        return self._session

    # internal query/search

    def _query(self, s: str, attributes: bool = False) -> List[JsonDict]:
        try:
            results = self.session().query(s).get("records", [])
        except SalesforceExpiredSession:
            logger.debug("Salesforce session expired")
            results = self.session(True).query(s).get("records", [])

        if not attributes:
            for row in results:
                del row["attributes"]

        return results

    def _query_all(self, s: str, flatten: bool = True) -> Iterator[JsonDict]:
        reconnect = True

        try:
            for row in self.session().query_all_iter(s):
                if flatten:
                    yield _flatten_row(row)
                else:
                    yield row
                reconnect = False
        except SalesforceExpiredSession:
            logger.debug("Salesforce session expired")
            if reconnect:
                for row in self.session(True).query_all_iter(s):
                    if flatten:
                        yield _flatten_row(row)
                    else:
                        yield row
            else:
                raise RuntimeError("Cannot reconnect Salesforce session after partial query")

    def _search(self, s):
        try:
            return self.session().search(s)
        except SalesforceExpiredSession:
            logger.debug("Salesforce session expired")
            return self.session(True).search(s)

    def rest_post(self, endpoint: str, params: dict) -> dict:
        sess = self.session()
        headers = {"Authorization": "Bearer " + sess.session_id}

        r = requests.post("https://" + sess.sf_instance + endpoint, json=params, headers=headers, timeout=self.timeout)
        r.raise_for_status()
        return r.json()

    def rest_get(self, endpoint: str) -> dict:
        sess = self.session()
        headers = {"Authorization": "Bearer " + sess.session_id}

        r = requests.get("https://" + sess.sf_instance + endpoint, headers=headers, timeout=self.timeout)
        r.raise_for_status()
        return r.json()

    # actual methods

    def query(self, query_str: str) -> List[JsonDict]:
        return self._query(query_str)

    def _get_one_field(self, results, name):
        return [row[name] for row in results]

    def get_all_objects(self) -> List[JsonDict]:
        query_str = "SELECT QualifiedApiName, Label FROM EntityDefinition ORDER BY QualifiedApiName"

        return self._query(query_str)

    def get_all_fields(self, object_name: str) -> List[JsonDict]:
        """Retrieves all fields of `object_name`.
        Warning: `object_name` is not escaped!
        """

        query_str = f"SELECT QualifiedApiName, DataType, Label FROM FieldDefinition WHERE EntityDefinition.QualifiedApiName = '{object_name}'"  # nosec

        return self._query(query_str)

    def search_fields(self, s: str, object_name: str, object_fields: Iterable[str]) -> dict:
        """Searches for `s` in object_name with object_fields.
        Warning: `object_name` and `object_fields` are not escaped!
        """

        s = sosl_escape(s)
        return self._search("FIND {{{}}} RETURNING {}({})".format(s, object_name, ", ".join(object_fields)))

    @staticmethod
    def dictkeymapper(it: Iterable[MappingT[str, str]], columns: MappingT[str, T]) -> Iterator[Dict[T, str]]:
        for row in it:
            yield {v: row.get(k, "") for k, v in columns.items()}

    def soql_to_pandas(
        self,
        query_str: str,
        columns: Optional[Union[SequenceT[str], MappingT[str, str]]] = None,
        progress: Optional[Progress] = None,
    ) -> "pd.DataFrame":
        import pandas as pd  # noqa: F811

        progress = progress or Progress()
        it = progress.track(self._query_all(query_str, flatten=True))

        if not columns:
            fieldnames: Optional[Iterable[str]] = None
        elif isinstance(columns, Sequence):
            fieldnames = columns
        elif isinstance(columns, Mapping):
            fieldnames = columns.keys()
        else:
            raise ValueError()

        df = pd.DataFrame.from_records(it, columns=fieldnames)
        if isinstance(columns, Mapping):
            df.columns = columns.values()
        return df

    def dump_csv(
        self,
        query_str: str,
        path: str,
        safe: bool = False,
        columns: Optional[Union[SequenceT[str], MappingT[str, str]]] = None,
        progress: Optional[Progress] = None,
    ) -> int:
        """Run SOQL `query_str` and dump results to csv file `path`.
        Returns the number of exported rows.
        """

        i = 0

        with sopen(path, "wt", encoding="utf-8", newline="", safe=safe) as csvfile:
            progress = progress or Progress()
            it = progress.track(self._query_all(query_str, flatten=True))

            if isinstance(columns, Mapping):
                it = self.dictkeymapper(it, columns)

            row = next(it)
            if not columns:
                fieldnames = row.keys()
            elif isinstance(columns, Sequence):
                fieldnames = columns
            elif isinstance(columns, Mapping):
                fieldnames = columns.values()
            else:
                raise ValueError()

            csvwriter = csv.DictWriter(csvfile, fieldnames, extrasaction="ignore")

            csvwriter.writeheader()
            csvwriter.writerow(row)
            i += 1

            for row in it:
                csvwriter.writerow(row)
                i += 1

        return i


class LiveAgentBase(Generic[ReturnTGet, ReturnTPost]):
    # https://help.salesforce.com/articleView?id=000331168&type=1&mode=1
    # https://help.salesforce.com/articleView?id=000340657&type=1&mode=1

    api_version = "42"

    def __init__(
        self,
        hostname: str,
        organization_id: str,
        deployment_id: str,
        button_id: str,
        scheme: str = "https",
        timeout: int = 30,
    ) -> None:
        self.hostname = hostname
        self.organization_id = organization_id
        self.deployment_id = deployment_id
        self.button_id = button_id
        self.scheme = scheme
        self.timeout = timeout

        if not (hostname and organization_id and deployment_id and button_id and scheme):
            raise ValueError("All arguments must have a value")

        self.key: Optional[str] = None
        self.affinity_token: Optional[str] = None
        self._sequence = 0
        self.last_offset = 0
        self.client_poll_timeout: Optional[int] = None

    # abstract

    def get_request(
        self, endpoint: str, headers: JsonDict, params: Optional[JsonDict] = None, timeout: Optional[float] = None
    ) -> ReturnTGet:
        raise NotImplementedError

    def post_request(
        self, endpoint: str, headers: JsonDict, json: Optional[JsonDict] = None, timeout: Optional[float] = None
    ) -> ReturnTPost:
        raise NotImplementedError

    # helper

    def urljoin(self, endpoint: str) -> str:
        return self.scheme + "://" + self.hostname + endpoint

    @property
    def sequence(self) -> str:
        self._sequence += 1
        return str(self._sequence)

    @staticmethod
    def _make_prechat_details(slots: Union[Dict[str, str], List[JsonDict]]) -> List[JsonDict]:
        if isinstance(slots, dict):
            prechat_details = []

            for label, value in slots.items():
                custom_detail = {
                    "label": label,
                    "value": value,
                    "displayToAgent": True,
                    "transcriptFields": [],
                    "entityMaps": [],
                }
                prechat_details.append(custom_detail)

        elif isinstance(slots, list):
            try:
                for custom_detail in slots:
                    custom_detail["label"]
                    custom_detail["value"]
                    custom_detail.setdefault("displayToAgent", True)
                    custom_detail.setdefault("transcriptFields", [])
                    custom_detail.setdefault("entityMaps", [])
            except KeyError as e:
                raise ValueError(f"slots is missing fields: {e}")

            prechat_details = slots

        else:
            raise TypeError("either prechat_details must be given or slots must be dict or list")

        return prechat_details

    def _availibility(self, res: JsonDict) -> bool:
        # this is not a staticmethod as to include potentially useful stack-info in logs

        for msg in res["messages"]:
            if msg["type"] == "Availability":
                return msg["message"]["results"][0].get("isAvailable", False)
            else:
                logger.error("Got unexpected message: %s", msg)

        raise SalesforceError("No Availability message in live agent response")

    def _reset(self, d: JsonDict) -> None:
        msg = d["messages"][0]
        assert msg["type"] == "ReconnectSession"

        self.affinity_token = msg["affinityToken"]
        reset = msg.get("resetSequence", True)
        if reset:
            self._sequence = 0

    def _set_session_info(self, response: JsonDict) -> None:
        self.key = response["key"]
        self.affinity_token = response["affinityToken"]
        self.client_poll_timeout = response["clientPollTimeout"]

    # high level

    def send(self, text: str) -> ReturnTPost:
        """Send text message to agent."""

        assert self.key
        assert self.affinity_token
        return self.rest_chat_message(self.key, self.affinity_token, text)

    def close(self) -> ReturnTPost:
        """Close live chat.
        Can raise a 403 HTTP error in case the agent already ended the chat.
        """

        assert self.key
        assert self.affinity_token
        return self.rest_chat_end(self.key, self.affinity_token)

    # low level

    def rest_session_id(self) -> ReturnTGet:
        endpoint = "/chat/rest/System/SessionId"

        headers = {
            "X-LIVEAGENT-AFFINITY": "null",
        }

        return self.get_request(endpoint, headers)

    def rest_chasitor_init(
        self,
        key: str,
        affinity_token: str,
        session_id: str,
        visitor_name: str,
        user_agent: str = "",
        language: str = "en-US",
        screen_resolution: str = "1920x1080",
        prechat_details: Optional[List[JsonDict]] = None,
        prechat_entities: Optional[List[JsonDict]] = None,
        receive_queue_updates: bool = True,
        is_post: bool = True,
    ) -> ReturnTPost:
        endpoint = "/chat/rest/Chasitor/ChasitorInit"

        headers = {
            "X-LIVEAGENT-AFFINITY": affinity_token,
            "X-LIVEAGENT-SESSION-KEY": key,
        }

        params = {
            "organizationId": self.organization_id,
            "deploymentId": self.deployment_id,
            "buttonId": self.button_id,
            "sessionId": session_id,
            "userAgent": user_agent,
            "language": language,
            "screenResolution": screen_resolution,
            "visitorName": visitor_name,
            "prechatDetails": prechat_details or [],
            "prechatEntities": prechat_entities or [],
            "receiveQueueUpdates": receive_queue_updates,
            "isPost": is_post,
        }

        return self.post_request(endpoint, headers, json=params)

    def rest_reconnect_session(self, key: str, affinity_token: str, offset: int) -> ReturnTGet:
        endpoint = "/chat/rest/System/ReconnectSession"

        headers = {
            "X-LIVEAGENT-AFFINITY": affinity_token,
            "X-LIVEAGENT-SESSION-KEY": key,
        }

        params = {"ReconnectSession.offset": offset}

        return self.get_request(endpoint, headers, params=params)

    def rest_chasitor_resync_state(self, key: str, affinity_token: str) -> ReturnTPost:
        endpoint = "/chat/rest/Chasitor/ChasitorResyncState"

        headers = {
            "X-LIVEAGENT-AFFINITY": affinity_token,
            "X-LIVEAGENT-SESSION-KEY": key,
        }

        params = {"organizationId": self.organization_id}

        return self.post_request(endpoint, headers, json=params)

    def rest_messages(self, key: str, affinity_token: str) -> ReturnTGet:
        endpoint = "/chat/rest/System/Messages"

        headers = {
            "X-LIVEAGENT-AFFINITY": affinity_token,
            "X-LIVEAGENT-SESSION-KEY": key,
        }

        return self.get_request(endpoint, headers, timeout=self.client_poll_timeout)

    def rest_chat_message(self, key: str, affinity_token: str, text: str) -> ReturnTPost:
        endpoint = "/chat/rest/Chasitor/ChatMessage"

        headers = {
            "X-LIVEAGENT-AFFINITY": affinity_token,
            "X-LIVEAGENT-SESSION-KEY": key,
        }

        params = {
            "text": text,
        }

        return self.post_request(endpoint, headers, json=params)

    def rest_chasitor_sneak_peek(self, key: str, affinity_token: str, position: int, text: str) -> ReturnTPost:
        endpoint = "/chat/rest/Chasitor/ChasitorSneakPeek"

        headers = {
            "X-LIVEAGENT-AFFINITY": affinity_token,
            "X-LIVEAGENT-SESSION-KEY": key,
        }

        params = {"position": position, "text": text}

        return self.post_request(endpoint, headers, json=params)

    def rest_chat_end(self, key: str, affinity_token: str) -> ReturnTPost:
        endpoint = "/chat/rest/Chasitor/ChatEnd"

        headers = {
            "X-LIVEAGENT-AFFINITY": affinity_token,
            "X-LIVEAGENT-SESSION-KEY": key,
        }

        params = {"type": "ChatEndReason", "reason": "client"}

        return self.post_request(endpoint, headers, json=params)

    def rest_availability(self, estimated_wait_time: bool = False) -> ReturnTGet:
        endpoint = "/chat/rest/Visitor/Availability"

        headers: JsonDict = {}

        params = {
            "org_id": self.organization_id,
            "deployment_id": self.deployment_id,
            # this should be an array according to docs. however arrays in url query params are not clearly defined.
            # `requests` repeats the key with different values, but it's not clear that this is the correct way.
            # edit: according to https://www.srinivas4sfdc.com/2019/12/live-agent-chat-rest-api-to-check.html
            # this should be a comma separated list.
            "Availability.ids": self.button_id,
            "Availability.needEstimatedWaitTime": str(int(estimated_wait_time)),
        }

        return self.get_request(endpoint, headers, params=params)


class LiveAgent(LiveAgentBase[JsonDict, bytes]):
    def get_request(
        self, endpoint: str, headers: JsonDict, params: Optional[JsonDict] = None, timeout: Optional[float] = None
    ) -> JsonDict:
        headers.setdefault("X-LIVEAGENT-API-VERSION", self.api_version)
        timeout = timeout or self.timeout

        r = requests.get(self.urljoin(endpoint), params=params, headers=headers, timeout=self.timeout)
        r.raise_for_status()
        if r.status_code == 204:
            return {}
        else:
            return r.json()

    def post_request(
        self, endpoint: str, headers: JsonDict, json: Optional[JsonDict] = None, timeout: Optional[float] = None
    ) -> bytes:
        headers.setdefault("X-LIVEAGENT-API-VERSION", self.api_version)
        headers.setdefault("X-LIVEAGENT-SEQUENCE", self.sequence)
        timeout = timeout or self.timeout

        r = requests.post(self.urljoin(endpoint), json=json, headers=headers, timeout=self.timeout)
        r.raise_for_status()
        return r.content

    # high level

    def connect(
        self,
        visitor_name: str,
        slots: Optional[Union[Dict[str, str], List[JsonDict]]] = None,
        prechat_details: Optional[List[JsonDict]] = None,
        prechat_entities: Optional[List[JsonDict]] = None,
    ) -> None:
        """Connect using `visitor_name` as name."""

        self._sequence = 0
        response = self.rest_session_id()

        self._set_session_info(response)
        session_id = response["id"]
        if not prechat_details and slots:
            prechat_details = self._make_prechat_details(slots)

        assert self.key
        assert self.affinity_token
        self.rest_chasitor_init(
            self.key,
            self.affinity_token,
            session_id,
            visitor_name,
            prechat_details=prechat_details,
            prechat_entities=prechat_entities,
        )

    def reconnect(self) -> None:
        """Use `reconnect()` whenever a 503 error is encounted."""

        assert self.key
        assert self.last_offset
        d = self.rest_reconnect_session(self.key, "null", self.last_offset)
        self._reset(d)
        assert self.affinity_token
        self.rest_chasitor_resync_state(self.key, self.affinity_token)

    def is_available(self) -> bool:
        """Check for agent availability."""

        res = self.rest_availability()
        return self._availibility(res)

    def wait_available(self, wait: float = 10) -> None:
        """Waits until an agent is available.
        Polls every `wait` seconds.
        """

        while True:
            if self.is_available():
                return

            time.sleep(wait)

    def receive(self, wait_forever: bool = False) -> List[JsonDict]:
        """Long poll messages."""

        assert self.key
        assert self.affinity_token
        while True:
            d = self.rest_messages(self.key, self.affinity_token)
            if not d:
                if wait_forever:
                    continue
                else:
                    return []

            self.last_offset = d.get("offset", 0)
            return d["messages"]


class LiveAgentAsync(LiveAgentBase[Coroutine[Any, Any, JsonDict], Coroutine[Any, Any, bytes]]):
    trust_env = True

    async def get_request(
        self, endpoint: str, headers: JsonDict, params: Optional[JsonDict] = None, timeout: Optional[float] = None
    ) -> JsonDict:
        headers = headers or {}
        headers.setdefault("X-LIVEAGENT-API-VERSION", self.api_version)
        timeout = timeout or self.timeout

        async with aiohttp.ClientSession(trust_env=self.trust_env) as session:
            async with session.get(self.urljoin(endpoint), headers=headers, params=params, timeout=timeout) as r:
                r.raise_for_status()
                if r.status == 204:
                    return {}
                else:
                    return await r.json()

    async def post_request(
        self, endpoint: str, headers: JsonDict, json: Optional[JsonDict] = None, timeout: Optional[float] = None
    ) -> bytes:
        headers = headers or {}
        headers.setdefault("X-LIVEAGENT-API-VERSION", self.api_version)
        headers.setdefault("X-LIVEAGENT-SEQUENCE", self.sequence)
        timeout = timeout or self.timeout

        async with aiohttp.ClientSession(trust_env=self.trust_env) as session:
            async with session.post(self.urljoin(endpoint), headers=headers, json=json, timeout=timeout) as r:
                r.raise_for_status()
                return await r.read()

    # high level

    async def connect(
        self,
        visitor_name: str,
        slots: Optional[Union[Dict[str, str], List[JsonDict]]] = None,
        prechat_details: Optional[List[JsonDict]] = None,
        prechat_entities: Optional[List[JsonDict]] = None,
    ) -> None:
        """Connect using `visitor_name` as name."""

        self._sequence = 0
        response = await self.rest_session_id()

        self._set_session_info(response)
        session_id = response["id"]
        if not prechat_details and slots:
            prechat_details = self._make_prechat_details(slots)

        assert self.key
        assert self.affinity_token
        await self.rest_chasitor_init(
            self.key,
            self.affinity_token,
            session_id,
            visitor_name,
            prechat_details=prechat_details,
            prechat_entities=prechat_entities,
        )

    async def reconnect(self) -> None:
        """Use `reconnect()` whenever a 503 error is encounted."""

        assert self.key
        assert self.last_offset
        d = await self.rest_reconnect_session(self.key, "null", self.last_offset)
        self._reset(d)
        assert self.affinity_token
        await self.rest_chasitor_resync_state(self.key, self.affinity_token)

    async def is_available(self) -> bool:
        """Check for agent availability."""

        res = await self.rest_availability()
        return self._availibility(res)

    async def wait_available(self, wait: float = 10) -> None:
        """Waits until an agent is available.
        Polls every `wait` seconds.
        """

        while True:
            if await self.is_available():
                return

            await asyncio.sleep(wait)

    async def receive(self, wait_forever: bool = False) -> List[JsonDict]:
        """Long poll messages."""
        assert self.key
        assert self.affinity_token

        while True:
            d = await self.rest_messages(self.key, self.affinity_token)
            if not d:
                if wait_forever:
                    continue
                else:
                    return []

            self.last_offset = d.get("offset", 0)
            return d["messages"]



from typing import Any, Iterable, Iterator, Optional, Sequence, TypeVar, Union

from typing_extensions import Self

from tqdm import tqdm

from .callbacks import BaseTask
from .callbacks import Progress as _Progress
from .callbacks import _Default

T = TypeVar("T")


class Task(BaseTask):
    def __init__(self, total: Optional[float] = None, description: Optional[float] = None, **kwargs: Any) -> None:
        self.pbar = tqdm(desc=description, total=total, **kwargs)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args):
        self.pbar.close()

    def advance(self, delta: float) -> None:
        self.pbar.update(n=delta)

    def update(
        self,
        *,
        completed: Optional[float] = _Default,
        total: Optional[float] = _Default,
        description: Optional[str] = _Default,
        **kwargs: Any,
    ) -> None:
        if completed is not _Default:
            self.pbar.n = completed
        if total is not _Default:
            self.pbar.total = total
        if description is not _Default:
            self.pbar.set_description(description)


class Progress(_Progress):
    def track(
        self,
        sequence: Union[Iterable[T], Sequence[T]],
        total: Optional[float] = None,
        description: Optional[str] = None,
        **kwargs: Any,
    ) -> Iterator[T]:
        return tqdm(sequence, description, total, **kwargs)

    def task(self, total: Optional[float] = None, description: Optional[str] = None, **kwargs: Any):
        return Task(total, description, **kwargs)

    def set_prolog(self, prolog: str) -> None:
        raise NotImplementedError

    def set_epilog(self, epilog: str) -> None:
        raise NotImplementedError

    def print(self, s: str, end="\n") -> None:
        tqdm.write(s, end=end)



import csv
from itertools import islice

import re2 as re

from .time import PrintStatementTime

# UNFINISHED!!! use re2 or hyperscan for pathological regex


class Browscap:
    def __init__(self, path="browscap.csv"):
        self.path = path

    @staticmethod
    def convert_to_regex(pattern):
        ret = f"^{re.escape(pattern)}$"  # fnmatch.translate(pattern)
        return ret.replace("\\?", ".").replace("\\*", ".*")

    def iter_patterns(self):
        with open(self.path, newline="") as fr:
            csvreader = csv.reader(fr)
            for row in islice(csvreader, 2, None):
                yield self.convert_to_regex(row[0]), row[6]

    def match_v1(self, useragent):
        data = list((re.compile(p, re.IGNORECASE), t) for p, t in self.iter_patterns())

        with PrintStatementTime():
            for pattern, type in data:
                if pattern.match(useragent):
                    print(type)

    def match_v2(self, useragent):
        pattern = "|".join(p for p, t in self.iter_patterns())
        pattern = re.compile(pattern, re.IGNORECASE, max_mem=1024**3)

        with PrintStatementTime():
            if pattern.match(useragent):
                print(True)


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("path", help="browscap.csv file")
    args = parser.parse_args()

    cap = Browscap(args.path)
    # cap.match_v1("Firefox")
    cap.match_v2("Firefox")



from typing import AbstractSet, Any, Dict, Optional, Sequence, Tuple, Type, TypeVar, Union

T = TypeVar("T")
U = TypeVar("U")


class NotFound(LookupError):
    """Raised when a search doesn't turn up any results.
    Similar to KeyError or IndexError.
    """


class ParseError(Exception):
    """Raised when the content to be parsed is malformed.
    Not a value error because usually the error is in some external resource
    which we don't have any control over.
    """

    def __init__(self, *args, data=None):
        super().__init__(*args)
        self.data = data


class DownloadFailed(Exception):
    """Raised when a download from an external resource by HTTP or any other protocol did not
    complete successfully. Could be raised anywhere between the initial TCP connection to
    a final integrity check.
    """


class AccessDenied(Exception):
    cached: Optional[bool]


class TemporaryError(Exception):
    """Signals that an failed operation can be retried in the future
    with a possibly different outcome.
    """


class IteratorExhausted(Exception):
    """Tried to get data from an exhausted Iterator."""


class WouldBlockForever(Exception):
    """Raised when a blocking operation would wait forever."""


# results, control flow


class ControlFlowException(Exception):
    """Not really an error, rather used for control flow handling
    where other constructs like `return` or `break` are not convenient.
    """


class Skip(ControlFlowException):
    pass


class NoResult(ControlFlowException):
    """Raised when an operation which maybe returns a result yields no result."""


class Break(ControlFlowException):
    """Control flow exception to break out of a recursive call."""


class NoActionNeeded(ControlFlowException):
    """Raised when nothing needs to be done / nothing was modified."""


# external errors


class ExternalError(Exception):
    """Raised when an error occurs due to an external resource which we don't have control over,
    or are not the only actor to control it.
    """


# aka ConsistencyError
class InconsistentState(ExternalError):
    """This is either raised when some external resource changed without our knowledge,
    or when our code violates some consistency assumptions.
    """


class ExternalProcedureUnavailable(ExternalError):
    """If some external resource cannot be reached by API or RPC or is otherwise not able to handle
    the request, this is raised. It is not raised for invalid requests to the resource which are
    correctly handled (or rather correctly not handled in thie case).
    """


class DatabaseUnavailable(ExternalError):
    """Raised when the connection to a database fails because the database cannot be reached or
    raises and error related to the database itself and not to the query.
    """


# runtime, possible coding errors


class ClosedObjectUsed(RuntimeError):
    def __init__(self, obj):
        RuntimeError.__init__(self, f"{obj.__class__.__name__} is already closed")


# values, input errors


class EmptyIterable(ValueError):
    """Raised when Iterable is passed which doesn't yield any values,
    and thus not resulted can be computed.
    """


class MalformedFile(ValueError):
    """Raised when a malformed File is passes as input.
    In contrast to `ParseError` this is usually a file we have control over.
    """


def assert_choice(name: str, value: Optional[T], choices: AbstractSet[T], optional: bool = False) -> None:
    if optional and value is None:
        return

    if value not in choices:
        raise ValueError("{} must be one of {}".format(name, ", ".join(map(str, choices))))


def assert_choices(name: str, values: Optional[Sequence[T]], choices: AbstractSet[T], optional: bool = False) -> None:
    if values is None:
        if optional:
            return
        else:
            raise TypeError("`None cannot be passed for `values` when `optional=False`")

    if frozenset(values) - choices:
        raise ValueError("{} must be a subset of {}".format(name, ", ".join(map(str, choices))))


def assert_choice_map(name: str, value: T, choices: Dict[T, U]) -> U:
    try:
        return choices[value]
    except KeyError:
        raise ValueError("{} must be one of {}".format(name, ", ".join(map(str, choices.keys()))))


def assert_type(name: str, value: Any, types: Union[Type[Any], Tuple[Type[Any], ...]]) -> None:
    if not isinstance(value, types):
        if not isinstance(types, tuple):
            types = (types,)
        raise TypeError(
            "{} must be one of these types: {}. Not: {}".format(name, ", ".join(map(str, types)), type(value))
        )


def assert_true(name: str, value) -> None:
    if not value:
        raise ValueError(f"{name} must be set")



from enum import Enum
from typing import Callable, Optional, Sequence, TypeVar

from .func import identity

T = TypeVar("T")
U = TypeVar("U")


class BisectRetVal(Enum):
    Lower = -1
    Equal = 0
    Higher = 1


def make_binary_search(target: int) -> Callable[[int], BisectRetVal]:
    def binary_search_func(x):
        if x < target:
            return BisectRetVal.Higher
        elif x > target:
            return BisectRetVal.Lower
        else:
            return BisectRetVal.Equal

    return binary_search_func


def make_binary_search_sequence(
    seq: Sequence[T], target: U, key: Optional[Callable[[T], U]] = None
) -> Callable[[int], BisectRetVal]:
    _key = key or identity

    def binary_search_func(x: int) -> BisectRetVal:
        result = _key(seq[x])
        if result < target:
            return BisectRetVal.Higher
        elif result > target:
            return BisectRetVal.Lower
        else:
            return BisectRetVal.Equal

    return binary_search_func


def bisect_left_generic(lo: int, hi: int, func: Callable[[int], BisectRetVal]) -> int:
    """Generic left bisection (binary search). Finds a specific `lo<=x<=hi` where
    `func(x) == BisectRetVal.Equal` or where the return value changes from BisectRetVal.Higher
    to `BisectRetVal.Lower`.
    """

    while lo < hi:
        mid = (lo + hi) // 2
        result = func(mid)
        if result == BisectRetVal.Higher:
            lo = mid + 1
        else:
            hi = mid
    return lo


def bisect_right_generic(lo: int, hi: int, func: Callable[[int], BisectRetVal]) -> int:
    """Generic right bisection (binary search). Finds a specific `lo<=x<=hi` where
    `func(x) == BisectRetVal.Equal` or where the return value changes from BisectRetVal.Higher
    to `BisectRetVal.Lower`.
    """

    while lo < hi:
        mid = (lo + hi) // 2
        result = func(mid)
        if result == BisectRetVal.Lower:
            hi = mid
        else:
            lo = mid + 1
    return lo


def bisect_left_sequence(seq: Sequence[T], target, key=None) -> int:
    """see: bisect.bisect_left"""

    func = make_binary_search_sequence(seq, target, key)
    return bisect_left_generic(0, len(seq), func)


def bisect_right_sequence(seq: Sequence[T], target, key=None) -> int:
    """see: bisect.bisect_right"""

    func = make_binary_search_sequence(seq, target, key)
    return bisect_right_generic(0, len(seq), func)


def search_sorted(seq: Sequence[T], newitem: U, key: Optional[Callable[[T], U]] = None) -> int:
    _key = key or identity

    for i, item in enumerate(seq):
        if newitem < _key(item):
            return i

    return len(seq)


if __name__ == "__main__":
    seq = [1, 2, 3, 4, 4, 5]
    from bisect import bisect_left, bisect_right

    for target in (1, 2, 3, 4, 5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5):
        print(bisect_left_sequence(seq, target) == bisect_left(seq, target))
        print(search_sorted(seq, target) == bisect_right_sequence(seq, target) == bisect_right(seq, target))



import concurrent.futures
import logging
import signal
import threading
from collections import deque
from concurrent.futures._base import FINISHED
from multiprocessing import Pool
from queue import Empty, Queue
from typing import Any, Callable, Deque, Dict, Generic, Iterable, Iterator, List, Optional, Set, Tuple, TypeVar, Union

from typing_extensions import Self

from .exceptions import NoResult, assert_choice

T = TypeVar("T")
U = TypeVar("U")
TaskT = Tuple[Callable, tuple, Dict[str, Any]]

logger = logging.getLogger(__name__)


class Worker(threading.Thread):
    TASK_IDLE = 0
    TASK_COMPLETE = 1
    TASK_EXCEPTION = 2

    running: bool

    def __init__(
        self,
        tasks: "Queue[Optional[Tuple[int, int, Callable, tuple, dict]]]",
        returns: "Queue[Tuple[int, int, Tuple[int, Union[Any, Exception]]]]",
    ):
        threading.Thread.__init__(self)
        self.tasks = tasks
        self.returns = returns

        self.daemon = True
        self.running = True
        self.start()

    def run(self) -> None:
        while self.running:
            # self.returns.put((self.TASK_IDLE, self))
            task = self.tasks.get(True)
            if task:
                state, id, func, args, kargs = task
            else:
                self.tasks.task_done()
                self.running = False
                break

            try:
                ret = func(*args, **kargs)
            except Exception as e:  # pylint: disable=broad-except
                logger.exception("threaded task failed")
                self.returns.put((state, self.TASK_EXCEPTION, (id, e)))
            else:
                self.returns.put((state, self.TASK_COMPLETE, (id, ret)))

            try:
                self.tasks.task_done()
            except ValueError:  # if the task was canceled inbetween
                pass

        logger.debug("thread ended")

    def cancel(self) -> None:
        self.running = False


class ThreadPool:
    tasks: "Queue[Optional[Tuple[int, int, Callable, tuple, dict]]]"
    returns: "Queue[Tuple[int, int, Tuple[int, Union[Any, Exception]]]]"
    state: int

    def __init__(self, num_threads: int) -> None:
        self.num_threads = num_threads
        self.tasks = Queue()
        self.returns = Queue()
        self.my_worker = [Worker(self.tasks, self.returns) for _ in range(num_threads)]
        self.state = 0  # this is used to be able to cancel (or ignore the results of ongoing) tasks

    def add_task(self, id: int, func: Callable[..., Any], *args: Any, **kargs: Any) -> None:
        self.tasks.put((self.state, id, func, args, kargs))

    def get(self) -> Tuple[int, Optional[Any]]:
        state: int
        type_: int
        result: Tuple[int, Union[Any, Exception]]

        while True:  # so invalid states can be ignored
            state, type_, result = self.returns.get(True)
            if state == self.state:
                if type_ == Worker.TASK_COMPLETE:
                    return result  # (id, ret)
                else:
                    id, e = result
                    return id, None  # maybe raise here?
            #  else: old result, ignore

    @staticmethod
    def clear(queue: "Queue[Any]") -> None:
        """Only undocumented methods."""

        with queue.mutex:
            queue.queue.clear()
            queue.all_tasks_done.notify_all()
            queue.unfinished_tasks = 0

    @staticmethod
    def _clear(queue: "Queue[Any]") -> None:
        """Only uses documented methods, but is slower."""

        while not queue.empty():
            queue.get()
            queue.task_done()

    def close(self, finishall: bool = False) -> None:
        """if `finishall` is True, it will finish all tasks currently in the queue,
        if `finishall` is False, only currently active tasks will be finished
        """

        if finishall:
            for _i in range(self.num_threads):
                self.tasks.put(None)
        else:
            for w in self.my_worker:
                w.cancel()

    def cancel(self) -> None:
        """Does not cancel already running tasks. Just clears queue and ignores
        results of currently running tasks.
        """

        self.clear(self.tasks)
        self.clear(self.returns)
        self.state += 1

    def wait(self) -> None:
        self.tasks.join()


def gather_all_unsorted(
    threadpool: ThreadPool, func: Callable, params: Iterable, *args: Any, **kwargs: Any
) -> Iterator[Tuple[Any, Any]]:
    """Runs multiple tasks concurrently and returns all results in the order
    of execution finish as soon as possible.
    """

    num_params = 0
    for i, param in enumerate(params):
        threadpool.add_task(i, func, param, *args, **kwargs)
        num_params = i + 1

    for _i in range(num_params):
        yield threadpool.get()


def gather_any(threadpool: ThreadPool, func: Callable, params: Iterable, *args: Any, **kwargs: Any) -> Tuple[Any, Any]:
    """Runs multiple tasks concurrently and just returns the result of the first
    task which finishes.
    """

    for id, result in gather_all_unsorted(threadpool, func, params, *args, **kwargs):
        threadpool.cancel()
        return id, result  # result will be None, if tasked threw exception
    raise NoResult("No task returned a valid result")


def NotThreadSafe(verify: bool = False) -> Callable[[type], type]:
    """Class decorator. Only works on new style classes (object)
    `verify=True` prohibits write access to attributes.
    """

    def OnClass(TheClass: type) -> type:
        if verify:
            orig_init = TheClass.__init__
            orig_setattr = TheClass.__setattr__

            def __setattr__(self, name, value):
                if self.thread_ident == threading.current_thread().ident:
                    orig_setattr(self, name, value)
                else:
                    raise RuntimeError(f"{TheClass.__name__}.{name} = {value} called from different thread")

            def __init__(self, *args, **kws):
                orig_setattr(self, "thread_ident", threading.current_thread().ident)
                orig_init(self, *args, **kws)

            TheClass.__init__ = __init__
            TheClass.__setattr__ = __setattr__
        return TheClass

    return OnClass


class AbortIteration(Exception):
    pass


class IterWorker(threading.Thread):
    STATE_WAITING = 0
    STATE_RUNNING = 1
    STATE_PAUSING = 2
    STATE_PAUSED = 3
    STATE_STOPPED = 4

    CMD_RESUME = 0
    CMD_ABORT = 1

    def __init__(self, taskqueue: "Queue[Optional[Iterable]]", onstatechange: Optional[Callable] = None):
        threading.Thread.__init__(self)
        self.queue = taskqueue
        self.onstatechange = onstatechange
        self.control: "Queue[int]" = Queue()
        self.state = self.STATE_WAITING

    @property
    def state(self) -> int:
        return self._state

    @state.setter
    def state(self, value: int) -> None:
        assert_choice("value", value, (0, 1, 2, 3, 4))

        self._state = value
        if self.onstatechange:
            self.onstatechange(value)

    def run(self) -> None:
        while True:
            iter = self.queue.get()
            if iter is None:
                self.state = self.STATE_STOPPED
                return
            if self.state == self.STATE_WAITING:
                self.state = self.STATE_RUNNING
            try:
                for __ in iter:
                    if self.state == self.STATE_PAUSING:
                        self.state = self.STATE_PAUSED
                        cmd = self.control.get()
                        if cmd == self.CMD_RESUME:
                            self.state = self.STATE_RUNNING
                            continue
                        elif cmd == self.CMD_ABORT:
                            break
            except AbortIteration:
                logger.warning("Task aborted")
            except Exception:
                logger.exception("Task threw exception")
            self.queue.task_done()  # untested
            self.state = self.STATE_WAITING

    def pause(self) -> None:
        if self.state == self.STATE_RUNNING:
            self.state = self.STATE_PAUSING

    def resume(self) -> None:
        if self.state == self.STATE_PAUSED:
            self.control.put(self.CMD_RESUME)

    def stop(self) -> None:
        if self.state < self.STATE_STOPPED:
            self.queue.put(None)

    def clear(self) -> None:
        try:
            while True:
                self.queue.get_nowait()
                self.queue.task_done()
        except Empty:
            pass

    def abort(self) -> None:
        self.pause()
        self.control.put(self.CMD_ABORT)


# was: DownloadWorker
class ProgressWorker(threading.Thread):
    running: bool
    task: Optional[TaskT]

    def __init__(self, manager: "ProgressThreadPool", tasks: "Queue[Optional[TaskT]]"):
        threading.Thread.__init__(self)
        self.manager = manager
        self.tasks = tasks

        self.running = True
        self.task = None
        self.done = None
        self.total = None

    def run(self):
        while self.running:
            self.task = self.tasks.get()
            if self.task is None:
                break
            func, args, kwargs = self.task

            try:
                result = func(self.report, *args, **kwargs)
            except Exception:
                logger.exception("Task failed")
                result = None
            self.task, self.done, self.total = None, None, None
            self.manager.finalize(result)

    def quit(self) -> None:
        self.running = False

    def report(self, done, total) -> None:
        self.done = done
        self.total = total

    def get(self) -> Optional[Tuple[TaskT, Any, Any]]:  # not using locks atm, can return inconsistent data
        if self.task and self.done and self.total:
            return (self.task, self.done, self.total)
        else:
            return None


# was: ThreadedDownloader
class ProgressThreadPool:
    completed: List[Any]
    failed: List[Tuple[Exception, Any]]

    def __init__(self, concurrent: int = 1) -> None:
        self.waiting_queue = Queue()
        self.completed = []
        self.failed = []
        self.lock = threading.Lock()

        self.workers = list(ProgressWorker(self, self.waiting_queue) for i in range(concurrent))
        for w in self.workers:
            w.daemon = True  # program will end even if threads are still running
            w.start()

    def stop(self) -> None:
        for _ in range(len(self.workers)):
            self.waiting_queue.put(None)

    def join(self, timeout: Optional[float] = None) -> bool:
        """Returns True if all workers were terminated, otherwise False"""

        for worker in self.workers:
            worker.join(timeout)

        self.workers = [worker for worker in self.workers if worker.is_alive()]

        return len(self.workers) == 0

    def start(self, callable: Callable, *args, **kwargs) -> None:  # todo: add optional task id here.
        self.waiting_queue.put((callable, args, kwargs))

    def finalize(self, result: Optional[Tuple[Callable, Optional[Exception], Any]]) -> None:
        if result is not None:
            setter, status, ret = result
            with self.lock:
                if status:
                    self.failed.append((status, ret))
                else:
                    self.completed.append(ret)
                    setter(ret)
        else:  # task failed
            pass

    def clear_completed(self) -> None:
        self.completed = []  # assignment is atomic right?

    def cancel_pending(self) -> None:
        # copied from https://stackoverflow.com/a/18873213
        with self.waiting_queue.mutex:
            self.waiting_queue.queue.clear()
            self.waiting_queue.all_tasks_done.notify_all()
            self.waiting_queue.unfinished_tasks = 0

    def get_waiting(self) -> List[TaskT]:
        with self.waiting_queue.mutex:
            return list(task for task in self.waiting_queue.queue if task)

    def get_completed(self) -> List[Any]:
        return self.completed

    def get_failed(self) -> List[Tuple[Exception, Any]]:
        return self.failed

    def get_running(self):  # no locks, inconsistent data
        return list(task for task in (w.get() for w in self.workers) if task)


class BoundedIterator:
    # similar: pip install bounded-iterator

    """Semaphor bounded queue. Can be used with `multiprocessing.Pool` for example.
    `imap()` calls iterable from same process but different thread.

    Other semaphore classes like `multiprocessing.BoundedSemaphore` can be used as well.
    """

    def __init__(
        self,
        size: int,
        it: Iterable[T],
        timeout: Optional[float] = None,
        semaphore: Callable[[int], Any] = threading.BoundedSemaphore,
    ) -> None:
        self.semaphore = semaphore(size)
        self.iterable = it
        self.timeout = timeout
        self.iterator: Optional[Iterator[T]] = None
        self.stopped = False

    def __iter__(self) -> Iterator[T]:
        self.iterator = iter(self.iterable)
        return self

    def __next__(self) -> T:
        if self.iterator is None:
            raise TypeError("iter not called")

        while True:
            if self.stopped:
                raise StopIteration("interrupted")
            if self.semaphore.acquire(timeout=self.timeout):
                break

        return next(self.iterator)

    def done(self) -> None:
        self.semaphore.release()

    def stop(self) -> None:
        self.stopped = True


class BufferedIterable(Generic[T]):
    def __init__(self, it: Iterable[T], bufsize: int):
        self.iterable = it
        self.iterator = None
        self.buffer: Deque[T] = deque([])
        self.bufsize = bufsize

    def __iter__(self) -> Iterator[T]:
        self.iterator = iter(self.iterable)
        return self

    def __next__(self) -> T:
        if self.iterator is None:
            raise TypeError

        try:
            while len(self.buffer) < self.bufsize:
                item = next(self.iterator)
                self.buffer.append(item)
        except StopIteration:
            pass

        try:
            return self.buffer.popleft()
        except IndexError:
            raise StopIteration


class CompletedFutures:
    def __init__(self, it: Iterable[concurrent.futures.Future], bufsize: int, timeout=None):
        self.iterable = it
        self.iterator: Optional[Iterator[concurrent.futures.Future]] = None
        self.futures: Set[concurrent.futures.Future] = set()
        self.bufsize = bufsize
        self.timeout = timeout

    def __iter__(self) -> Iterator[concurrent.futures.Future]:
        self.iterator = iter(self.iterable)
        return self

    def __next__(self) -> concurrent.futures.Future:
        if self.iterator is None:
            raise TypeError

        try:
            while len(self.futures) < self.bufsize:
                item = next(self.iterator)
                self.futures.add(item)
        except StopIteration:
            pass

        done, not_done = concurrent.futures.wait(
            self.futures, timeout=self.timeout, return_when=concurrent.futures.FIRST_COMPLETED
        )
        if done:
            ret = done.pop()
            self.futures = done | not_done
            return ret

        raise StopIteration

    def cancel_pending(self):
        for future in self.futures:
            future.cancel()


def _ignore_sigint() -> None:
    """This need to be pickle'able to work with `multiprocessing.Pool`."""

    try:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
    except ValueError:  # ignore for threadpools as 'signal only works in main thread'
        pass


def parallel_map(
    func: Callable[[T], U],
    it: Iterable[T],
    poolcls: Optional[Callable] = None,
    ordered: bool = True,
    parallel: bool = True,
    workers: Optional[int] = None,
    bufsize: int = 1,
    chunksize: int = 1,
) -> Iterator[U]:
    """Parallel map which uses multiprocessing to distribute tasks.
    `bufsize` should be used to limit memory usage when the iterable `it`
    can be processed faster than the output iterator is consumed.

    Keyboard interrupts are ignored in child processes. They should however be terminated
    correctly by the main thread.
    """

    if parallel:
        if poolcls is None:
            poolcls = Pool

        q = BoundedIterator(bufsize, it, timeout=1)

        initializer = _ignore_sigint if type(poolcls) is Pool else None

        with poolcls(workers, initializer) as p:
            if ordered:
                process = p.imap
            else:
                process = p.imap_unordered

            try:
                for item in process(func, q, chunksize):
                    yield item
                    q.done()
            except GeneratorExit:
                logging.warning("interrupted")
                # q.timeout = 1  # does this help?
                q.done()  # the semaphore in the bounded queue might block otherwise
                q.stop()
                raise
    else:
        yield from map(func, it)


def FutureWithResult(result):
    future = concurrent.futures.Future()
    future._result = result
    future._state = FINISHED
    return future


def executor_map(
    func: Callable[[T], U],
    it: Iterable[T],
    executercls: Optional[Callable] = None,
    ordered: bool = True,
    parallel: bool = True,
    workers: Optional[int] = None,
    bufsize: int = 1,
) -> Iterator[concurrent.futures.Future]:
    """Starts processing when the iterator is started to be consumed."""

    if parallel:
        if executercls is None:
            executercls = concurrent.futures.ThreadPoolExecutor

        def futures() -> Iterator[concurrent.futures.Future]:
            for item in it:
                yield executor.submit(func, item)

        with executercls(workers) as executor:
            if ordered:
                bufit: Iterable = BufferedIterable(futures(), bufsize + executor._max_workers)
            else:
                bufit = CompletedFutures(futures(), bufsize + executor._max_workers)

            yield from bufit

    else:
        yield from map(FutureWithResult, map(func, it))


class ThreadsafeList(list):  # untested!!!
    """This is a list object with context manager to handle locking to perform multiple operations
    on a list in a threadsafe way.
    Example.
    l = ThreadsafeList([1, 2, 3])
    l.append(4) # atomic, no locking needed
    with l:
            l[0] += 4 # not atomic, thus executed with context manager
    """

    def __init__(self, it) -> None:
        list.__init__(self, it)
        self.lock = threading.Lock()

    def __enter__(self) -> Self:
        self.lock.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.lock.release()


def idsleeprandom(i):
    import random
    import time

    time.sleep(random.random())  # nosec
    return i


def benchmark():
    """According to these benchmark results, executor_map is faster for threads,
    whereas parallel_map is faster for processes.
    """

    from multiprocessing.pool import ThreadPool

    from genutility.func import identity
    from genutility.iter import consume
    from genutility.time import PrintStatementTime

    executercls = concurrent.futures.ThreadPoolExecutor
    with PrintStatementTime("executor_map(ThreadPool) ordered=True took {delta}s"):
        consume(executor_map(identity, range(100000), executercls=executercls, ordered=True, parallel=True))
    with PrintStatementTime("executor_map(ThreadPool) ordered=False took {delta}s"):
        consume(executor_map(identity, range(100000), executercls=executercls, ordered=False, parallel=True))

    executercls = concurrent.futures.ProcessPoolExecutor
    with PrintStatementTime("executor_map(ProcessPool) ordered=True took {delta}s"):
        consume(executor_map(identity, range(10000), executercls=executercls, ordered=True, parallel=True))
    with PrintStatementTime("executor_map(ProcessPool) ordered=False took {delta}s"):
        consume(executor_map(identity, range(10000), executercls=executercls, ordered=False, parallel=True))

    poolcls = ThreadPool
    with PrintStatementTime("parallel_map(ThreadPool) ordered=True took {delta}s"):
        consume(parallel_map(identity, range(100000), poolcls=poolcls, ordered=True, parallel=True))
    with PrintStatementTime("parallel_map(ThreadPool) ordered=False took {delta}s"):
        consume(parallel_map(identity, range(100000), poolcls=poolcls, ordered=False, parallel=True))

    poolcls = Pool
    with PrintStatementTime("parallel_map(ProcessPool) ordered=True took {delta}s"):
        consume(parallel_map(identity, range(10000), poolcls=poolcls, ordered=True, parallel=True))
    with PrintStatementTime("parallel_map(ProcessPool) ordered=False took {delta}s"):
        consume(parallel_map(identity, range(10000), poolcls=poolcls, ordered=False, parallel=True))


if __name__ == "__main__":
    benchmark()



import asyncio
import logging
from collections import deque
from datetime import datetime
from typing import Any, Deque, List, Optional, Set

import aiohttp

from .datetime import now
from .url import get_filename_from_url

logger = logging.getLogger(__name__)


class DownloadTask:
    def __init__(self, url: str, path: str = ".") -> None:
        self.url = url
        self.path = path
        self.downloaded = 0
        self.resumable = False
        self.dt_started: Optional[datetime] = None
        self.dt_finished = now()

    def __hash__(self):
        return hash((self.url, self.path))

    def start(self) -> None:
        logger.info("starting download")
        self.dt_started = now()

    def done(self) -> None:
        logger.info("finished download")
        self.dt_finished = now()


class DownloadManager:
    def __init__(self, concurrent_downloads: int = 3) -> None:
        self.loop = asyncio.get_event_loop()
        self.timeout = aiohttp.ClientTimeout(total=None, sock_read=60)
        self.session = aiohttp.ClientSession(loop=self.loop, timeout=self.timeout, auto_decompress=False)
        self.concurrent_downloads = concurrent_downloads
        # self.sem = asyncio.Semaphore(1000)
        self.chunksize = 1024 * 1024  # file write buffer

        self.queue: Deque[DownloadTask] = deque()
        self.active: Set[DownloadTask] = set()
        self.done: List[DownloadTask] = []
        self.error: List[DownloadTask] = []

    def status(self) -> str:
        total_active = sum(t.downloaded for t in self.active)
        total_done = sum(t.downloaded for t in self.done)
        total_error = sum(t.downloaded for t in self.error)

        return "Queued: {}, active: {}, done: {}, error: {}\nDownload active: {}, done: {}, error: {}".format(
            len(self.queue), len(self.active), len(self.done), len(self.error), total_active, total_done, total_error
        )

    def _enqueue(self, task: DownloadTask, priority: Any) -> None:
        self.queue.append(task)

    def _start(self, task: DownloadTask) -> Optional[asyncio.Task]:
        if task not in self.active:
            self.active.add(task)
            atask = asyncio.ensure_future(self._download(task))
            return atask
        else:
            self.error.append(task)
            return None

    def _trystart(self) -> Optional[asyncio.Task]:
        if len(self.active) < self.concurrent_downloads:
            try:
                task = self.queue.popleft()
                return self._start(task)
            except IndexError:
                if not self.active:
                    logger.info("all done")
                    # self.loop.stop()
                    # task = asyncio.ensure_future(self._close())

        return None

    async def _download(self, task: DownloadTask) -> None:
        task.start()
        # await asyncio.sleep(10)

        # send http head request first to check for range support

        try:
            # async with self.session.get(task.url, headers={"Range": "bytes=0-10"}) as response:
            async with self.session.get(task.url, headers={}) as response:
                stream = response.content
                try:
                    size: Optional[int] = int(response.headers.get("content-length", ""))
                except (ValueError, TypeError):
                    size = None

                accept_range = response.headers.get("Accept-Ranges", "none").lower()

                if response.status == 200:  # range not supported
                    pass
                elif response.status == 206:  # range supported
                    if accept_range != "bytes":
                        raise RuntimeError("Only bytes content ranges are supported")
                    bytes_range = response.headers.get("Content-Range")  # 'bytes 0-10/46239'
                    raise RuntimeError(f"Range requests are not supported yet: {bytes_range}")

                with open(task.path, "wb", buffering=self.chunksize) as fw:
                    async for data in stream.iter_any():
                        task.downloaded += len(data)
                        fw.write(data)

                if size and size != task.downloaded:
                    print("incomplete", task.downloaded, "of", size)

        except asyncio.TimeoutError:
            self.error.append(task)
        else:
            self.done.append(task)

        task.done()
        self.active.remove(task)
        self._trystart()

    def download(
        self, url: str, path: Optional[str] = None, priority: int = 0, force: bool = False
    ) -> Optional[asyncio.Task]:
        path = path or get_filename_from_url(url)
        logger.info("Starting download <%s> to <%s>", url, path)
        task = DownloadTask(url, path)
        if force:
            return self._start(task)
        else:
            self._enqueue(task, priority)
            return self._trystart()

    async def _close(self) -> None:
        await self.session.close()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import wx
    from wxasync import AsyncBind, StartCoroutine, WxAsyncApp

    DOWNLOAD_URL = "http://releases.ubuntu.com/18.04.3/ubuntu-18.04.3-live-server-amd64.iso"

    class TestFrame(wx.Frame):
        def __init__(self, parent=None):
            super().__init__(parent)
            self.dm = DownloadManager()

            vbox = wx.BoxSizer(wx.VERTICAL)
            button1 = wx.Button(self, label="Submit")
            self.edit = wx.StaticText(self, style=wx.ALIGN_CENTRE_HORIZONTAL | wx.ST_NO_AUTORESIZE)
            self.edit_timer = wx.StaticText(self, style=wx.ALIGN_CENTRE_HORIZONTAL | wx.ST_NO_AUTORESIZE)
            vbox.Add(button1, 2, wx.EXPAND | wx.ALL)
            vbox.AddStretchSpacer(1)
            vbox.Add(self.edit, 1, wx.EXPAND | wx.ALL)
            vbox.Add(self.edit_timer, 1, wx.EXPAND | wx.ALL)
            self.SetSizer(vbox)
            self.Layout()
            AsyncBind(wx.EVT_BUTTON, self.async_callback, button1)
            StartCoroutine(self.update_clock, self)

        async def async_callback(self, event):
            self.edit.SetLabel("WX WAITING")
            self.dm.download(DOWNLOAD_URL)
            self.edit.SetLabel("WX COMPLETE")

        async def update_clock(self):
            while True:
                self.edit_timer.SetLabel(self.dm.status())
                await asyncio.sleep(0.5)

    async def main():
        app = WxAsyncApp()
        frame = TestFrame()
        frame.Show()
        app.SetTopWindow(frame)
        await app.MainLoop()

    loop = asyncio.events.new_event_loop()
    try:
        asyncio.events.set_event_loop(loop)
        loop.run_until_complete(main())
    finally:
        asyncio.events.set_event_loop(None)
        loop.close()



import logging
import os
import subprocess  # nosec
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from ..os import CurrentWorkingDirectory
from ..string import surrounding_join
from ..subprocess import force_decode  # nosec

logger = logging.getLogger(__name__)


class RarError(Exception):
    returncode: int
    cmd: str
    output: str

    def __init__(self, msg: str, returncode: int, cmd: str, output: str) -> None:
        Exception.__init__(self, msg)
        self.returncode = returncode
        self.cmd = cmd
        self.output = output


class WinRarExitCodes(Enum):
    SUCCESS = 0  # Successful operation
    WARNING = 1  # Non fatal error(s) occurred
    FATAL_ERROR = 2  # A fatal error occurred
    CRC_ERROR = 3  # Invalid checksum, data is damaged
    LOCKED_ARCHIVE = 4  # Attempt to modify a locked archive
    WRITE_ERROR = 5  # Write error
    OPEN_ERROR = 6  # Open file error
    USER_ERROR = 7  # Wrong command line option
    MEMORY_ERROR = 8  # Not enough memory
    CREATE_ERROR = 9  # Create file error
    NO_FILES = 10  # No files matching the specified mask and options were found
    BAD_PASS = 11  # Wrong password.
    USER_BREAK = 255  # User stopped the process


@dataclass
class Flag:
    value: bool
    name: str


@dataclass
class Option:
    value: Union[bool, int, str]
    template: str


class Rar:
    windows_executable = Path("C:/Program Files/WinRAR/Rar.exe")
    flags: Dict[str, Flag]
    options: Dict[str, Option]

    def __init__(self, archive: Path, executable: Optional[Path] = None) -> None:
        """archive: archive to work with, everything which is supported by winrar
        executable: path/to/Rar.exe"""

        self.exe = executable or self.windows_executable

        if not self.exe.is_file():
            raise ValueError("Invalid executable")

        self.archive = archive
        self.filelist: List[str] = []
        self.cmd = ""
        self.flags = {}
        self.flags["lock"] = Flag(False, "k")
        self.flags["delete"] = Flag(False, "df")
        self.flags["test"] = Flag(False, "t")
        self.flags["filetime"] = Flag(False, "tl")
        self.flags["append_archive_name"] = Flag(False, "ad")
        self.options = {}
        self.options["split"] = Option(False, "v%s")
        self.options["compression"] = Option(False, "m%u")
        self.options["password"] = Option(False, "p%s")
        self.options["password_header"] = Option(False, "hp%s")
        self.options["recovery"] = Option(False, "rr%up")
        self.options["recovery_volumes"] = Option(False, "rr%u%%")

    def add_file(self, pathname: str) -> None:
        self.filelist.append(pathname)

    def add_files(self, filelist: Iterable[str]) -> None:
        self.filelist.extend(filelist)

    def set_compression(self, level: Union[int, bool]) -> None:
        """level: 0 store, 1 fastest, 2 fast, 3 normal, 4 good, 5 best (default: 3)"""

        if level not in range(0, 6) and level is not False:
            raise ValueError("Invalid parameter: Set compression level (0-store...3-default...5-best)")
        self.options["compression"].value = level

    def set_password(self, password, encrypt_filenames: bool = False) -> None:
        if encrypt_filenames:
            self.options["password_header"].value = password
            self.options["password"].value = False
        else:
            self.options["password"].value = password
            self.options["password_header"].value = False

    def add_recovery_info(self, rr: int) -> None:
        """rr: recovery record in percent, only 1-10 is valid"""

        if rr < 1 or rr > 10:
            raise ValueError("Only 1%-10% valid")
        self.options["recovery"].value = rr

    def add_recovery_volumes(self, rv: int) -> None:
        """rv: number of recovery volumes"""

        self.options["recovery_volumes"].value = rv

    def split(self, split: str):
        self.options["split"].value = split

    def lock(self, flag: bool = True) -> None:
        self.flags["lock"].value = flag

    def delete_after_archiving(self, flag: bool = True) -> None:
        """flag (bool): delete files after archiving"""

        self.flags["delete"].value = flag

    def test_after_archiving(self, flag: bool = True) -> None:
        """flag (bool): test archive after archiving"""

        self.flags["test"].value = flag

    def set_archive_to_filetime(self, flag: bool = True) -> None:
        self.flags["filetime"].value = flag

    def commandline(self, cmd: str) -> None:
        self.cmd = cmd.strip()

    def _execute(self, args: str) -> str:
        cmd = f"{self.exe} {args}"
        try:
            ret = subprocess.check_output(cmd, stderr=subprocess.STDOUT, cwd=os.getcwd())  # nosec
        except subprocess.CalledProcessError as e:
            output = force_decode(e.output)
            raise RarError(
                f"Calling `{e.cmd}` failed with error code {e.returncode}", e.returncode, e.cmd, output
            )  # should use only stderr

        return force_decode(ret)

    def test(self, password: str = "-") -> None:  # nosec
        self._execute(f't -p{password} "{self.archive}"')

    def get_files_str(self) -> str:
        return surrounding_join(" ", self.filelist, '"', '"')

    def get_flag_str(self) -> str:
        return surrounding_join(" ", [value.name for value in self.flags.values() if value.value is True], "-", "")

    def get_options_str(self) -> str:
        return surrounding_join(
            " ", [value.template % value.value for value in self.options.values() if value.value is not False], "-", ""
        )

    def _get_args(self, command: str) -> str:
        return f'{command} {self.get_flag_str()} {self.get_options_str()} {self.cmd} "{self.archive}" {self.get_files_str()}'

    def create(self) -> None:
        self._execute(self._get_args("a"))

    def extract(self, directory: str, mode: int = 0) -> None:
        self.filelist = [directory]
        if mode == 1:
            self.flags["append_archive_name"].value = True
        self._execute(self._get_args("x"))

    def close(self) -> None:
        pass


def create_rar_from_folder(
    path: Path,
    dest_path: Optional[Path] = None,
    profile_setter_func: Optional[Callable] = None,
    filter_func: Callable = lambda x: True,
    name_transform: Callable = lambda x: x,
) -> bool:
    if not path.is_dir():
        return False

    if dest_path is None:
        dest_path = path.parent

    with CurrentWorkingDirectory(path):
        try:
            r = Rar(dest_path / f"{name_transform(path.name)}.rar")
            if profile_setter_func:
                profile_setter_func(r)
            with os.scandir(".") as it:
                for entry in filter(filter_func, it):  # was: scandir_rec
                    r.add_file(entry.path)
            r.create()
        except RarError as e:
            logger.error(f"{str(e)}\n{e.output}")
            return False

    return True


_CWD_PATH = Path(".")


def create_rar_from_file(
    path: Path,
    dest_path: Path = _CWD_PATH,
    profile_setter_func: Optional[Callable[[Rar], Any]] = None,
    name_transform: Callable = lambda x: x,
) -> bool:
    if dest_path == ".":
        dest_path = path.parent

    with CurrentWorkingDirectory(path.parent):
        try:
            r = Rar(dest_path / f"{name_transform(path.name)}.rar")
            if profile_setter_func:
                profile_setter_func(r)
            r.add_file(path.name)
            r.create()
        except RarError as e:
            logger.error(f"{str(e)}\n{e.output}")
            return False

    return True



import logging
import os
import re
import struct
import warnings
from base64 import b64decode
from collections import namedtuple
from pathlib import Path
from typing import Any, Collection, Dict, Iterable, Iterator, List, NamedTuple, Optional, Tuple, Union

import pkg_resources

from ..csv import iter_csv
from ..exceptions import Break, ParseError
from ..file import BufferedBinaryIoT, read_or_raise
from ..iter import batch

# http://mp4ra.org/#/atoms
# https://wiki.multimedia.cx/index.php/QuickTime_container#QuickTime_Atom_Reference
# https://sno.phy.queensu.ca/~phil/exiftool/TagNames/QuickTime.html

""" mp4notes
    `stsd` can be both, be a parent and a leaf atom.
    `meta` can be both, be a parent and a leaf atom.
"""


class BoxParser:
    def __init__(self, fin: BufferedBinaryIoT, size: Optional[int] = None):
        self.fin = fin
        self.size = size
        self.delta = 0

    def unpack(self, format: str, size: int):
        ret = struct.unpack(format, read_or_raise(self.fin, size))
        self.delta += size
        return ret

    @staticmethod
    def read_c_string(fin: BufferedBinaryIoT, size: int) -> bytes:
        ret: List[bytes] = []
        while len(ret) < size:
            c = fin.read(1)
            ret.append(c)
            if c == b"\0":
                break
        return b"".join(ret)

    def c_string(self, encoding: Optional[str] = None) -> Union[bytes, str]:
        assert self.size is not None, "Size not specified"
        s = self.read_c_string(self.fin, self.size - self.delta)
        self.delta += len(s)
        s = s.rstrip(b"\0")
        if encoding is None:
            return s
        else:
            try:
                return s.decode(encoding)
            except UnicodeDecodeError:
                ret = s.decode("latin-1")  # should never fail
                logging.warning("'%s' is not a valid %s string", ret, encoding)
                return ret


def named_batch(entries: Iterable, length: int, named_tuple_cls: object) -> List[tuple]:
    assert len(named_tuple_cls._fields) == length, "length parameter doesn't match named tuple size"
    return list(batch(entries, length, named_tuple_cls._make))


# named tuples


class SampleToChunkEntry(NamedTuple):
    first_chunk: int
    samples_per_chunk: int
    sample_description_index: int


class CompositionOffsetEntry(NamedTuple):
    sample_count: int
    sample_offset: int


class TimeToSampleEntry(NamedTuple):
    sample_count: int
    sample_delta: int


class FilePartitionEntry(NamedTuple):
    block_count: int
    block_size: int


ItemLocationEntryWithIndex = namedtuple(
    "ItemLocationEntryWithIndex", ["extent_index", "extent_offset", "extent_length"]
)
ItemLocationEntry = namedtuple("ItemLocationEntry", ["extent_offset", "extent_length"])

# atoms


def stco(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    """ChunkOffsetBox"""

    assert version == 0, "Version not 0"
    p = BoxParser(fin, size)

    (entry_count,) = p.unpack(">L", 4)
    chunk_offsets = p.unpack(f">{entry_count}L", entry_count * 4)

    return {"chunk_offsets": chunk_offsets}, p.delta


def fpar(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    """FilePartitionBox"""

    assert version in (0, 1), "Version not 0 or 1"
    p = BoxParser(fin, size)

    if version == 0:
        (item_ID,) = p.unpack(">H", 2)
    elif version == 1:
        (item_ID,) = p.unpack(">L", 4)

    (
        packet_payload_size,
        reserved,
        FEC_encoding_ID,
        FEC_instance_ID,
        max_source_block_length,
        encoding_symbol_length,
        max_number_of_encoding_symbols,
    ) = p.unpack(">H2B4H", 12)
    assert reserved == 0

    scheme_specific_info = b64decode(p.c_string("ascii"))
    if version == 0:
        (entry_count,) = p.unpack(">H", 2)
    elif version == 1:
        (entry_count,) = p.unpack(">L", 4)

    entries = p.unpack(f">{'HL'*entry_count}", entry_count * 6)
    return {
        "item_ID": item_ID,
        "packet_payload_size": packet_payload_size,
        "FEC_encoding_ID": FEC_encoding_ID,
        "FEC_instance_ID": FEC_instance_ID,
        "max_source_block_length": max_source_block_length,
        "encoding_symbol_length": encoding_symbol_length,
        "max_number_of_encoding_symbols": max_number_of_encoding_symbols,
        "scheme_specific_info": scheme_specific_info,
        "file_partition_entries": named_batch(entries, 2, FilePartitionEntry),
    }, p.delta


def mfhd(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    """MovieFragmentHeaderBox"""

    assert version == 0
    p = BoxParser(fin, size)

    (sequence_number,) = p.unpack(">L", 4)

    return {"sequence_number": sequence_number}, p.delta


def co64(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    assert version == 0
    p = BoxParser(fin, size)

    (entry_count,) = p.unpack(">L", 4)
    chunk_offsets = p.unpack(f">{entry_count}Q", entry_count * 8)

    return {"chunk_offsets": chunk_offsets}, p.delta


def prft(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    p = BoxParser(fin, size)

    if version == 0:
        reference_track_ID, ntp_timestamp, media_time = p.unpack(">LQL", 16)
    elif version == 1:
        reference_track_ID, ntp_timestamp, media_time = p.unpack(">LQQ", 20)
    else:
        assert False

    return {"reference_track_ID": reference_track_ID, "ntp_timestamp": ntp_timestamp, "media_time": media_time}, p.delta


def iref_parser(fin, size, version):
    p = BoxParser(fin, size)
    if version == 0:
        (from_item_id,) = p.unpack(">H", 2)
    elif version == 1:
        (from_item_id,) = p.unpack(">L", 4)
    else:
        assert False, f"Unsupported version: {version}"

    if size == p.delta:  # empty box, found in test files, is this standard?
        to_item_ids = ()
    else:
        (to_item_ids_num,) = p.unpack(">H", 2)

        if version == 0:
            to_item_ids = p.unpack(f">{to_item_ids_num}H", to_item_ids_num * 2)
        elif version == 1:
            to_item_ids = p.unpack(f">{to_item_ids_num}L", to_item_ids_num * 4)
        else:
            assert False, f"Unsupported version: {version}"

    return {"from_item_id": from_item_id, "to_item_ids": to_item_ids}, p.delta


def dimg(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    """Not a fullbox, inherits parent version"""

    return iref_parser(fin, size, version)


def thmb(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    """Not a fullbox, inherits parent version"""

    return iref_parser(fin, size, version)


def cdsc(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    """Not a fullbox, inherits parent version"""

    return iref_parser(fin, size, version)


def ctts(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    """Composition Time to Sample Box / CompositionOffsetBox"""
    p = BoxParser(fin, size)

    (entry_count,) = p.unpack(">L", 4)

    if version == 0:
        entries = p.unpack(f">{entry_count*2}L", entry_count * 2 * 4)
    elif version == 1:
        entries = p.unpack(f">{'Ll'*entry_count}", entry_count * 2 * 4)
    else:
        assert False, f"Unsupported version: {version}"

    return {"composition_offset_entries": named_batch(entries, 2, CompositionOffsetEntry)}, p.delta


def stsc(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    """Sample To Chunk Box"""

    assert version == 0, f"Unsupported version: {version}"
    p = BoxParser(fin, size)
    (entry_count,) = p.unpack(">L", 4)
    entries = p.unpack(f">{entry_count*3}L", entry_count * 3 * 4)
    return {"sample_to_chunk_entries": named_batch(entries, 3, SampleToChunkEntry)}, p.delta


def stts(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    """TimeToSampleBox"""

    assert version == 0, f"Unsupported version: {version}"
    p = BoxParser(fin, size)
    (entry_count,) = p.unpack(">L", 4)
    entries = p.unpack(f">{entry_count*2}L", entry_count * 2 * 4)
    return {"time_to_samples_entries": named_batch(entries, 2, TimeToSampleEntry)}, p.delta


def uuid(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    p = BoxParser(fin, size)
    (uuid,) = p.unpack(">16s", 16)
    return {"uuid": uuid}, p.delta


def ftyp(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    p = BoxParser(fin, size)
    major_brand, minor_version = p.unpack(">4sL", 8)
    return {"major_brand": major_brand, "minor_version": minor_version}, p.delta


def stsd(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    assert version == 0, f"Unsupported version: {version}"
    p = BoxParser(fin, size)
    (entry_count,) = p.unpack(">L", 4)
    return {"entry_count": entry_count}, p.delta


def url(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    assert version == 0, f"Unsupported version: {version}"
    p = BoxParser(fin, size)
    url = p.c_string("utf-8")
    return {"url": url}, p.delta


def urn(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    assert version == 0, f"Unsupported version: {version}"
    p = BoxParser(fin, size)
    urn = p.c_string("utf-8")
    name = p.c_string("utf-8")
    return {"urn": urn, "name": name}, p.delta


def dref(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:  # needs to be parsed!!!
    p = BoxParser(fin, size)
    (entry_count,) = p.unpack(">L", 4)
    return {"entry_count": entry_count}, p.delta


def iinf(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:  # needs to be parsed!!!
    p = BoxParser(fin, size)
    if version == 0:
        (entry_count,) = p.unpack(">H", 2)
    elif version == 1:
        (entry_count,) = p.unpack(">L", 4)
    else:
        assert False, f"Unsupported version: {version}"

    return {"entry_count": entry_count}, p.delta


def pitm(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    """Primary Item Box"""

    p = BoxParser(fin, size)
    if version == 0:
        (item_ID,) = p.unpack(">H", 2)
    elif version == 1:
        (item_ID,) = p.unpack(">L", 4)
    else:
        assert False, f"Unsupported version: {version}"

    return {"item_ID": item_ID}, p.delta


def hdlr(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    assert version == 0, f"Unsupported version: {version}"

    p = BoxParser(fin, size)
    pre_defined, handler_type, reserved, reserved, reserved = p.unpack(">L4sLLL", 20)
    # assert pre_defined == 0
    name = p.c_string("utf-8")
    return {"handler_type": handler_type, "name": name}, p.delta


def tfdt(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    p = BoxParser(fin, size)
    if version == 0:
        baseMediaDecodeTime = p.unpack(">L", 4)
    elif version == 1:
        baseMediaDecodeTime = p.unpack(">Q", 8)

    return {"baseMediaDecodeTime": baseMediaDecodeTime}, p.delta


def frma(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    p = BoxParser(fin, size)
    (data_format,) = p.unpack(">4s", 4)

    return {"data_format": data_format}, p.delta


def schm(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    p = BoxParser(fin, size)

    scheme_type, scheme_version = p.unpack(">4sL", 8)
    ret = {"scheme_type": scheme_type, "scheme_version": scheme_version}
    if int.from_bytes(flags, byteorder="big") & 0x000001:
        scheme_uri = p.c_string()
        ret["scheme_uri"] = scheme_uri

    return ret, p.delta


def infe(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    p = BoxParser(fin, size)
    ret = {}
    if version in (0, 1):
        item_ID, item_protection_index = p.unpack(">HH", 4)
        ret["item_ID"] = item_ID
        ret["item_protection_index"] = item_protection_index

    # if version == 1:
    #     unsigned int(32) extension_type; //optional
    #     ItemInfoExtension(extension_type); //optional

    if version >= 2:
        if version == 2:
            (item_ID,) = p.unpack(">H", 2)
        elif version == 3:
            (item_ID,) = p.unpack(">L", 4)
        ret["item_ID"] = item_ID

        item_protection_index, item_type = p.unpack(">H4s", 6)
        ret["item_protection_index"] = item_protection_index
        ret["item_type"] = item_type

        item_name = p.c_string("utf-8")
        ret["item_name"] = item_name
        if item_type == b"mime":
            content_type = p.c_string("utf-8")
            ret["content_type"] = content_type
            # string content_encoding; //optional
        elif item_type == b"uri":
            item_uri_type = p.c_string("utf-8")
            ret["item_uri_type"] = item_uri_type

    return ret, p.delta


def size_to_format_char(size: int) -> str:
    return {4: "L", 8: "Q"}[size]


def iloc(fin: BufferedBinaryIoT, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    p = BoxParser(fin, size)
    ret: Dict[str, Any] = {}

    # assert fin.tell() % 8 == 0

    a, b = p.unpack(">BB", 2)
    offset_size, length_size = divmod(a, 16)
    base_offset_size, index_size_reserved = divmod(b, 16)

    if version == 1 or version == 2:
        index_size = index_size_reserved
    else:
        reserved = index_size_reserved

    if version < 2:
        (item_count,) = p.unpack(">H", 2)
    elif version == 2:
        (item_count,) = p.unpack(">L", 4)

    ret["items"] = []

    for i in range(item_count):
        if version < 2:
            (item_ID,) = p.unpack(">H", 2)
        elif version == 2:
            (item_ID,) = p.unpack(">L", 4)

        if version == 1 or version == 2:
            (a,) = p.unpack(">H", 2)
            reserved, construction_method = divmod(a, 16)
        else:
            construction_method = None

        if base_offset_size == 0:
            data_reference_index, extent_count = p.unpack(">HH", 4)
            base_offset = None
        else:
            fmt_char = size_to_format_char(base_offset_size)
            data_reference_index, base_offset, extent_count = p.unpack(f">H{fmt_char}H", 4 + base_offset_size)

        if (version == 1 or version == 2) and index_size > 0:
            fmt_chars = (
                size_to_format_char(index_size) + size_to_format_char(offset_size) + size_to_format_char(length_size)
            )
            entries = p.unpack(
                f">{'{fmt_chars}'*extent_count}", (index_size + offset_size + length_size) * extent_count
            )
            entries = named_batch(entries, 3, ItemLocationEntryWithIndex)
        else:
            fmt_chars = size_to_format_char(offset_size) + size_to_format_char(length_size)
            entries = p.unpack(f">{fmt_chars*extent_count}", (offset_size + length_size) * extent_count)
            entries = named_batch(entries, 2, ItemLocationEntry)

        ret["items"].append(
            {
                "item_ID": item_ID,
                "construction_method": construction_method,
                "data_reference_index": data_reference_index,
                "base_offset": base_offset,
                "entries": entries,
            }
        )

    return ret, p.delta


parsers = {
    "stsd": stsd,
    "iinf": iinf,
    "infe": infe,
    "pitm": pitm,
    "dref": dref,
    "url ": url,
    "urn ": urn,
    "hdlr": hdlr,
    "ftyp": ftyp,
    "dimg": dimg,
    "prft": prft,
    "tfdt": tfdt,
    "frma": frma,
    "schm": schm,
    "co64": co64,
    "stco": stco,
    "uuid": uuid,
    "ctts": ctts,
    "stsc": stsc,
    "stts": stts,
    "mfhd": mfhd,
    "fpar": fpar,
    "thmb": thmb,
    "cdsc": cdsc,
    "iloc": iloc,
}

versions = {
    "iref": (0, 1),
    "meta": (0,),
    "elng": (0,),
    "stdp": (0,),
    "fecr": (0, 1),
}


def _load_atoms() -> Dict[str, Tuple[str, str, str]]:
    package_name = __package__ or "genutility"
    atoms_path = pkg_resources.resource_filename(package_name, "data/mp4-atoms.tsv")

    out: Dict[str, Tuple[str, str, str]] = {}

    try:
        for fourcc, type, boxtype, description, _ in iter_csv(atoms_path, delimiter="\t", skip=1):
            assert fourcc not in out, f"Invalid fourcc: {fourcc}"
            out[fourcc] = (type, boxtype, description)
    except ValueError:
        logging.exception("Failed to parse atoms file at line %s", len(out) + 1)
        raise

    return out


atoms = _load_atoms()
atomcodep = re.compile(rb"[0-9a-zA-Z ]{4}")  # what do the specs say here?


def parse_atom(fin: BufferedBinaryIoT, code: str, size: int, version: int, flags: bytes) -> Tuple[dict, int]:
    try:
        func = parsers[code]
        content, delta = func(fin, size, version, flags)
    except KeyError:
        content = {}
        delta = 0

    return content, delta


def read_atom(fin: BufferedBinaryIoT, parent_version: Optional[int] = None) -> Tuple[int, str, int, int, int, bytes]:
    pos = fin.tell()

    p = BoxParser(fin)

    size, code = p.unpack(">L4s", 8)

    if not atomcodep.match(code):
        raise ParseError(f"{code!r} @ {pos} is not a valid atom code")

    code = code.decode("ascii")  # cannot fail

    if size == 1:  # 64bit size
        (size,) = p.unpack(">Q", 8)

    try:
        boxtype = atoms[code][1]
    except KeyError:  # treat it as leaf and skip it
        boxtype = "box"
        warnings.warn(f"Unknown atom: '{code}'. Skipping...", stacklevel=2)

    if boxtype == "fullbox":
        version, flags = p.unpack(">B3s", 4)

        try:
            supported = versions[code]
            assert version in supported, "Version not supported"
        except KeyError:
            pass

    else:
        version = parent_version
        flags = b""

    return pos, code, size, p.delta, version, flags


def _enum_atoms(
    fin: BufferedBinaryIoT,
    total_size: int,
    depth: int,
    parse_atoms: bool = True,
    unparsed_data: bool = False,
    version: Optional[int] = None,
) -> Iterator[Tuple[int, int, str, int, Optional[dict], Optional[bytes]]]:
    while fin.tell() < total_size:
        pos, type, size, delta, version, flags = read_atom(fin, version)

        if parse_atoms:
            content, d = parse_atom(fin, type, size - delta, version, flags)
            delta += d
        else:
            content = None

        if size == 0:
            raise Break("Atom extends to the end of the file")  # just stop parsing here

        atom_end = pos + size

        try:
            boxtype = atoms[type][0]
        except KeyError:  # treat it as leaf and skip it
            boxtype = "leaf"
            warnings.warn(f"Unknown atom: '{type}'. Skipping...", stacklevel=2)

        if boxtype == "cont":
            yield depth, pos, type, size, content, None

            yield from _enum_atoms(fin, atom_end, depth + 1, parse_atoms, unparsed_data, version)
        elif boxtype == "leaf":
            if unparsed_data:
                leaf = fin.read(size - delta)
                yield depth, pos, type, size, content, leaf
            else:
                yield depth, pos, type, size, content, None
                fin.seek(atom_end, os.SEEK_SET)
        else:
            assert False, "Invalid boxtype"

    if fin.tell() != total_size:
        raise ParseError(f"Invalid file structure. Possibly truncated. {fin.tell()}/{total_size}")


def enumerate_atoms(
    path: str, parse_atoms: bool = False, unparsed_data: bool = False
) -> Iterator[Tuple[int, int, str, int, Optional[dict], Optional[bytes]]]:
    """Takes an ISO/IEC base media file format / mp4 file `path`
    and yields (depth, position, code, size, parsed_data, unparsed_data) tuples.
    Unknown atoms will print a warning.
    """

    total_size = os.path.getsize(path)
    with open(path, "rb") as fr:
        try:
            yield from _enum_atoms(fr, total_size, 0, parse_atoms, unparsed_data)
        except Break:
            pass
        except EOFError:
            raise ParseError("Truncated file.")


if __name__ == "__main__":
    from argparse import ArgumentParser
    from os import fspath
    from sys import stderr

    import pandas as pd
    from rich.progress import Progress as RichProgress

    from genutility.filesystem import scandir_ext
    from genutility.iter import list_except
    from genutility.rich import Progress

    atoms_path = pkg_resources.resource_filename(__package__, "data/mp4-atoms.tsv")
    df = pd.read_csv(atoms_path, sep="\t")
    df.sort_values("fourcc").to_csv(atoms_path + ".new", sep="\t", index=False)

    def bytes_from_ascii(s):
        return s.encode("ascii")

    parser = ArgumentParser()
    parser.add_argument("path", type=Path, help="Input file or directory")
    parser.add_argument("-e", "--errors-only", action="store_true")
    parser.add_argument("-r", "--recursive", action="store_true")
    parser.add_argument(
        "--extensions", nargs="+", default=[".mp4", ".mov", ".f4v", ".heif", ".heic", ".3gp", ".3g2", ".mj2"]
    )
    parser.add_argument(
        "--type",
        nargs="+",
        help="Limit output to following types, or in comination with --errors-only only log errors if last tag is doesn't have this type.",
    )
    parser.add_argument("--search", type=bytes_from_ascii)
    parser.add_argument("--no-parse-atoms", action="store_false")
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    def print_atoms(path: Path, parse_atoms: bool, only_type: Collection[str], search: bytes) -> None:
        unparsed_data = search or only_type

        for depth, pos, type, size, content, leaf in enumerate_atoms(
            fspath(path), parse_atoms=parse_atoms, unparsed_data=unparsed_data
        ):
            if only_type and type in only_type:
                print("--" * depth, pos, type, size, content, leaf)
            elif search and leaf and search in leaf:
                print("--" * depth, pos, type, size, content, search)
            else:
                leavsize = len(leaf) if leaf else 0
                print("--" * depth, pos, type, size, content, leavsize)

    def print_atoms_error_only(path: Path, parse_atoms: bool, only_type: Collection[str]) -> bool:
        exc, res = list_except(enumerate_atoms(fspath(path), parse_atoms=parse_atoms))
        if exc:
            if only_type is None or (res and res[-1][2] not in only_type):
                for depth, pos, type, size, _, _ in res:
                    print("--" * depth, pos, type, size, file=stderr)
                logging.exception("Enumerating atoms of %s failed", path, exc_info=exc)
                return True
        return False

    with RichProgress() as progress:
        p = Progress(progress)

        if args.path.is_file():
            if args.errors_only:
                print_atoms_error_only(args.path, args.no_parse_atoms, args.type)
            else:
                print_atoms(args.path, args.no_parse_atoms, args.type, args.search)

        elif args.path.is_dir():
            errors_count = 0
            total_count = 0
            for path in p.track(scandir_ext(args.path, args.extensions, rec=args.recursive)):
                if args.errors_only:
                    total_count += 1
                    errors_count += int(print_atoms_error_only(path, args.no_parse_atoms, args.type))
                else:
                    print(path)
                    print_atoms(path, args.no_parse_atoms, args.type, args.search)
                    print()

            if args.errors_only:
                print(f"{errors_count}/{total_count} files failed to parse")

        else:
            assert False
